(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure

$Id: map_add.v,v 1.3 2003/11/11 13:42:37 monniaux Exp $
*)

Fixpoint coqadd [tree : t] : key -> a -> t := [k : key] [data : a]
  Cases tree of
  | Empty => (Node Empty k data Empty (S O))
  | (Node l v d r h) =>
     Cases (compare k v) of
     | EQUAL => (Node l k data r h)
     | LESS => (bal (coqadd l k data) v d r)
     | GREATER => (bal l v d (coqadd r k data))
     end
  end.

Lemma coqadd_bounded_down : (k, k_add: key) (data_add : a)
  (Horder : (compare k_add k) = GREATER)
  (tree : t) (Hbound : (bounded_down k tree))
  (bounded_down k (coqadd tree k_add data_add)).
Proof.
  Intros k k_add data_add Horder tree.
  Elim tree.
  Intro _.
  Simpl.
  Apply node_bounded_down.
  Apply empty_bounded_down.

  Assumption.

  Apply empty_bounded_down.

  Intros l Hl k_node data_node r Hr h Hbound.
  Simpl.
  Inversion Hbound.
  Cut
   (compare k_add k_node)=LESS
   ->(bounded_down k
       Cases (compare k_add k_node) of
         LESS => (bal (coqadd l k_add data_add) k_node data_node r)
       | EQUAL => (Node l k_add data_add r h)
       | GREATER => (bal l k_node data_node (coqadd r k_add data_add))
       end).
  Cut
   (compare k_add k_node)=EQUAL
   ->(bounded_down k
       Cases (compare k_add k_node) of
         LESS => (bal (coqadd l k_add data_add) k_node data_node r)
       | EQUAL => (Node l k_add data_add r h)
       | GREATER => (bal l k_node data_node (coqadd r k_add data_add))
       end).
  Cut
   (compare k_add k_node)=GREATER
   ->(bounded_down k
       Cases (compare k_add k_node) of
         LESS => (bal (coqadd l k_add data_add) k_node data_node r)
       | EQUAL => (Node l k_add data_add r h)
       | GREATER => (bal l k_node data_node (coqadd r k_add data_add))
       end).
  Intros HO1 HO2 HO3.
  NewDestruct (compare k_add k_node).
  Apply HO3; Trivial.

  Apply HO2; Trivial.

  Apply HO1; Trivial.

  Intro Horder2.
  Replace (compare k_add k_node) with GREATER.
  Apply bal_bounded_down.
  Assumption.

  Assumption.

  Apply Hr.
  Assumption.

  Intro rew.
  Rewrite -> rew.
  Apply node_bounded_down.
  Assumption.

  Replace (compare k_add k) with (compare k_node k).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Assumption.

  Intro rew.
  Rewrite -> rew.
  Apply bal_bounded_down.
  Assumption.

  Apply Hl.
  Assumption.

  Assumption.
Qed.

Lemma coqadd_bounded_up : (k, k_add: key) (data_add : a)
  (Horder : (compare k_add k) = LESS)
  (tree : t) (Hbound : (bounded_up k tree))
  (bounded_up k (coqadd tree k_add data_add)).
Proof.
  Intros k k_add data_add Horder tree.
  Elim tree.
  Intro _.
  Simpl.
  Apply node_bounded_up.
  Apply empty_bounded_up.

  Assumption.

  Apply empty_bounded_up.

  Intros l Hl k_node data_node r Hr h Hbound.
  Simpl.
  Inversion Hbound.
  Cut
   (compare k_add k_node)=LESS
   ->(bounded_up k
       Cases (compare k_add k_node) of
         LESS => (bal (coqadd l k_add data_add) k_node data_node r)
       | EQUAL => (Node l k_add data_add r h)
       | GREATER => (bal l k_node data_node (coqadd r k_add data_add))
       end).
  Cut
   (compare k_add k_node)=EQUAL
   ->(bounded_up k
       Cases (compare k_add k_node) of
         LESS => (bal (coqadd l k_add data_add) k_node data_node r)
       | EQUAL => (Node l k_add data_add r h)
       | GREATER => (bal l k_node data_node (coqadd r k_add data_add))
       end).
  Cut
   (compare k_add k_node)=GREATER
   ->(bounded_up k
       Cases (compare k_add k_node) of
         LESS => (bal (coqadd l k_add data_add) k_node data_node r)
       | EQUAL => (Node l k_add data_add r h)
       | GREATER => (bal l k_node data_node (coqadd r k_add data_add))
       end).
  Intros HO1 HO2 HO3.
  NewDestruct (compare k_add k_node).
  Apply HO3; Trivial.

  Apply HO2; Trivial.

  Apply HO1; Trivial.

  Intro Horder2.
  Rewrite -> Horder2.
  Apply bal_bounded_up.
  Assumption.

  Assumption.

  Apply Hr.
  Assumption.

  Intro rew.
  Rewrite -> rew.
  Apply node_bounded_up.
  Assumption.

  Replace (compare k_add k) with (compare k_node k).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Assumption.

  Intro rew.
  Rewrite -> rew.
  Apply bal_bounded_up.
  Assumption.

  Apply Hl.
  Assumption.

  Assumption.
Qed.


Lemma coqadd_wf_order : (tree : t) (tree_wf_order : (wf_order tree))
  (k_add : key) (data_add : a)  (wf_order (coqadd tree k_add data_add)).
Proof.
  Intro tree.
  Elim tree.
  Simpl.
  Intros.
  Apply node_wf_order.
  Apply empty_wf_order.

  Apply empty_wf_order.

  Apply empty_bounded_up.

  Apply empty_bounded_down.

  Intros.
  Simpl.
  Cut
   (compare k_add k)=LESS
   ->(wf_order (bal (coqadd left k_add data_add) k data right)).
  Cut
   (compare k_add k)=GREATER
   ->(wf_order (bal left k data (coqadd right k_add data_add))).
  Cut
   (compare k_add k)=EQUAL
   ->(wf_order (Node left k_add data_add right height0)).
  Intros HO1 HO2 HO3.
  NewDestruct (compare k_add k).
  Apply HO3; Trivial.

  Apply HO1; Trivial.

  Apply HO2; Trivial.

  Intro HO1.
  Apply node_wf_order.
  Inversion tree_wf_order.
  Assumption.

  Inversion tree_wf_order.
  Assumption.

  Inversion tree_wf_order.
  Apply wf_order_overbound_equal with k:=k.
  Assumption.

  Assumption.

  Apply wf_order_underbound_equal with k:=k.
  Assumption.

  Inversion tree_wf_order.
  Assumption.

  Intro order.
  Apply bal_wf_order.
  Inversion tree_wf_order.
  Assumption.

  Apply H0.
  Inversion tree_wf_order.
  Assumption.

  Inversion tree_wf_order.
  Assumption.

  Inversion tree_wf_order.
  Apply coqadd_bounded_down.
  Assumption.

  Assumption.

  Intro Horder.
  Apply bal_wf_order.
  Apply H.
  Inversion tree_wf_order.
  Assumption.

  Inversion tree_wf_order.
  Assumption.

  Apply coqadd_bounded_up.
  Assumption.

  Inversion tree_wf_order.
  Assumption.

  Inversion tree_wf_order.
  Assumption.
Qed.

Definition func_add := [f : key -> (option a)]
  [k_add : key] [data_add : a] [k : key]
  Cases (compare k k_add) of
  | EQUAL => (Some a data_add)
  | _ => (f k)
  end.

Lemma find_add_imm : (ft : t) (ft_wf_order : (wf_order ft))
  (k_add : key) (data : a) (k_find : key)
  (equality : ((compare k_find k_add) = EQUAL))
  ((coqfind (coqadd ft k_add data) k_find) = (Some a data)).
Proof.
  Intro ft.
  Elim ft.
  Intro ft_wf_order.
  Simpl.
  Intros.
  Rewrite -> equality.
  Trivial.

  Intros l Hl k_node data_node r Hr h ft_wf_order k_add data_add k_find
   equality.
  Simpl.
  Cut
   (compare k_add k_node)=LESS
   ->(coqfind (bal (coqadd l k_add data_add) k_node data_node r) k_find)
      =(Some a data_add).
  Intro HR1.
  Cut
   (compare k_add k_node)=EQUAL
   ->(coqfind (Node l k_add data_add r h) k_find)=(Some a data_add).
  Intro HR2.
  Cut
   (compare k_add k_node)=GREATER
   ->(coqfind (bal l k_node data_node (coqadd r k_add data_add)) k_find)
      =(Some a data_add).
  Intro HR3.
  NewDestruct (compare k_add k_node).
  Apply HR1.
  Trivial.

  Apply HR2.
  Trivial.

  Apply HR3.
  Trivial.

  Clear HR1 HR2.
  Intro order.
  Rewrite -> find_bal_greater.
  Apply Hr.
  Inversion ft_wf_order.
  Assumption.

  Assumption.

  Inversion ft_wf_order.
  Assumption.

  Apply coqadd_wf_order.
  Inversion ft_wf_order.
  Assumption.

  Inversion ft_wf_order.
  Assumption.

  Apply coqadd_bounded_down.
  Assumption.

  Inversion ft_wf_order.
  Assumption.

  Replace (compare k_find k_node) with (compare k_add k_node).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Intro Horder.
  Simpl.
  Rewrite -> equality.
  Trivial.

  Intro Horder.
  Replace (Some a data_add) with
  (coqfind (coqadd l k_add data_add) k_find).
  Apply find_bal_less.
  Apply coqadd_wf_order.
  Inversion ft_wf_order.
  Assumption.

  Inversion ft_wf_order.
  Assumption.

  Apply coqadd_bounded_up.
  Assumption.

  Inversion ft_wf_order.
  Assumption.

  Inversion ft_wf_order.
  Assumption.

  Replace (compare k_find k_node) with (compare k_add k_node).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Apply Hl.
  Inversion ft_wf_order.
  Assumption.

  Assumption.  
Qed.

Theorem add_correct : (ft : t) (ft_wf_order : (wf_order ft))
  (k_add : key) (data : a) (k : key)
  ((func_add (coqfind ft) k_add data k) = (coqfind (coqadd ft k_add data) k)).
Proof.
  Intro.
  Elim ft.
  Intros.
  Unfold func_add.
  Simpl.
  Trivial.

  Intros l Hl k_node data_node r Hr h ft_wf_order k_add data_add k_find.
  Unfold func_add.
  Simpl.
  Inversion ft_wf_order.
  Cut
   (cmp1:compare_result)
     (compare k_find k_add)=cmp1
     ->(cmp2:compare_result)
         (compare k_add k_node)=cmp2
         ->Cases (compare k_find k_add) of
             LESS =>
              Cases (compare k_find k_node) of
                LESS => (coqfind l k_find)
              | EQUAL => (Some a data_node)
              | GREATER => (coqfind r k_find)
              end
           | EQUAL => (Some a data_add)
           | GREATER =>
              Cases (compare k_find k_node) of
                LESS => (coqfind l k_find)
              | EQUAL => (Some a data_node)
              | GREATER => (coqfind r k_find)
              end
           end
            =(coqfind
               Cases (compare k_add k_node) of
                 LESS =>
                  (bal (coqadd l k_add data_add) k_node data_node r)
               | EQUAL => (Node l k_add data_add r h)
               | GREATER =>
                  (bal l k_node data_node (coqadd r k_add data_add))
               end k_find).
  Intro Z.
  Apply Z with cmp1:=(compare k_find k_add) cmp2:=(compare k_add k_node);
   Trivial.

  Intro cmp1.
  NewDestruct cmp1.
  Intro HO1.
  Intro cmp2.
  NewDestruct cmp2.
  Intro HO2.
  Rewrite -> HO1.
  Rewrite -> HO2.
  Replace (compare k_find k_node) with LESS.
  Symmetry.
  Replace (coqfind l k_find) with
  (coqfind (coqadd l k_add data_add) k_find).
  Apply find_bal_less.
  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_up; Assumption.

  Assumption.

  Apply compare_trans1 with y:=k_add; Assumption.

  Rewrite <- Hl.
  Unfold func_add.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Symmetry.
  Apply compare_trans1 with y:=k_add; Assumption.

  Intro HO2.
  Rewrite -> HO1.
  Rewrite -> HO2.
  Replace (compare k_find k_node) with LESS.
  Simpl.
  Rewrite -> HO1.
  Trivial.

  Symmetry.
  Replace LESS with (compare k_find k_add).
  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  Intro HO2.
  Rewrite -> HO1.
  Rewrite -> HO2.
  Cut
   (cmp3:compare_result)
     (compare k_find k_node)=cmp3
     ->Cases (compare k_find k_node) of
         LESS => (coqfind l k_find)
       | EQUAL => (Some a data_node)
       | GREATER => (coqfind r k_find)
       end
        =(coqfind (bal l k_node data_node (coqadd r k_add data_add))
           k_find).
  Intro Z.
  Apply Z with cmp3:=(compare k_find k_node); Trivial.

  Intro cmp3.
  Intro HO3.
  NewDestruct cmp3.
  Rewrite -> HO3.
  Symmetry.
  Apply find_bal_less.
  Assumption.

  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_down; Assumption.

  Assumption.

  Rewrite -> HO3.
  Symmetry.
  Apply find_bal_imm.
  Assumption.

  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_down; Assumption.

  Assumption.

  Rewrite -> HO3.
  Symmetry.
  Replace (coqfind r k_find) with
  (coqfind (coqadd r k_add data_add) k_find).
  Apply find_bal_greater.
  Assumption.

  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_down; Assumption.

  Assumption.

  Rewrite <- Hr.
  Unfold func_add.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Intro HO1.
  Replace (compare k_find k_node) with (compare k_add k_node).
  Intros cmp2 HO2.
  NewDestruct cmp2.
  Rewrite -> HO1.
  Rewrite -> HO2.
  Symmetry.
  Rewrite -> find_bal_less.
  Rewrite <- Hl.
  Unfold func_add.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Apply coqadd_wf_order.
  Assumption.

  Assumption.

  Apply coqadd_bounded_up.
  Assumption.

  Assumption.

  Assumption.

  Replace (compare k_find k_node) with (compare k_add k_node).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Rewrite -> HO1.
  Rewrite -> HO2.
  Simpl.
  Replace (compare k_find k_add) with EQUAL.
  Trivial.

  Rewrite -> HO1.
  Rewrite -> HO2.
  Symmetry.
  Rewrite -> find_bal_greater.
  Rewrite <- Hr.
  Unfold func_add.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Assumption.

  Apply coqadd_wf_order.
  Assumption.

  Assumption.

  Apply coqadd_bounded_down.
  Assumption.

  Assumption.

  Replace (compare k_find k_node) with (compare k_add k_node).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Intro HO1.
  Rewrite -> HO1.
  Intros cmp2 HO2.
  NewDestruct cmp2.
  Rewrite -> HO2.
  Cut
   (cmp3:compare_result)
     (compare k_find k_node)=cmp3
     ->Cases (compare k_find k_node) of
         LESS => (coqfind l k_find)
       | EQUAL => (Some a data_node)
       | GREATER => (coqfind r k_find)
       end
        =(coqfind (bal (coqadd l k_add data_add) k_node data_node r)
           k_find).
  Intro Z.
  Apply Z with cmp3:=(compare k_find k_node).
  Trivial.

  Intros cmp3 HO3.
  NewDestruct cmp3.
  Rewrite -> HO3.
  Symmetry.
  Replace (coqfind l k_find) with
  (coqfind (coqadd l k_add data_add) k_find).
  Apply find_bal_less.
  Apply coqadd_wf_order.
  Assumption.

  Assumption.

  Apply coqadd_bounded_up.
  Assumption.

  Assumption.

  Assumption.

  Assumption.

  Rewrite <- Hl.
  Unfold func_add.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Rewrite -> HO3.
  Symmetry.
  Apply find_bal_imm.
  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_up; Assumption.

  Assumption.

  Assumption.

  Rewrite -> HO3.
  Symmetry.
  Apply find_bal_greater.
  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_up; Assumption.

  Assumption.

  Assumption.

  Rewrite -> HO2.
  Replace (compare k_find k_node) with GREATER.
  Simpl.
  Rewrite -> HO1.
  Trivial.

  Replace GREATER with (compare k_find k_add).
  Apply compare_equiv2.
  Assumption.

  Rewrite -> HO2.
  Replace (compare k_find k_node) with GREATER.
  Symmetry.
  Rewrite -> find_bal_greater.
  Rewrite <- Hr.
  Unfold func_add.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Assumption.

  Apply coqadd_wf_order; Assumption.

  Assumption.

  Apply coqadd_bounded_down; Assumption.

  Apply compare_reverse2.
  Apply compare_trans1 with y:=k_add.
  Apply compare_reverse1.
  Assumption.

  Apply compare_reverse1.
  Assumption.

  Symmetry.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=k_add.
  Apply compare_reverse1.
  Assumption.

  Apply compare_reverse1.
  Assumption.
Qed.

