(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Lemma compare_tactic_helper : (P : Prop) (comparison : compare_result)
  ((cmp : compare_result) comparison = cmp -> P) -> P.
Proof.
  Intros P comparison H.
  Apply H with cmp:=comparison.
  Trivial.
Qed.

Tactic Definition TCompare x y Horder :=
  Apply compare_tactic_helper with comparison := (compare x y);
  Intros compare_tactic_cmp Horder;
  Try Rewrite Horder;
  NewDestruct compare_tactic_cmp;
  Idtac.

Lemma compare_equiv2 : (x, y, z : key)
  ((compare y z) = EQUAL) -> ((compare x y) = (compare x z)).
Proof.
  Intros x y z He.
  Replace (compare x y) with (invert_compare_result (compare y x)).
  Replace (compare x z) with (invert_compare_result (compare z x)).
  Replace (compare z x) with (compare y x).
  Trivial.

  Apply compare_equiv.
  Assumption.

  Symmetry.
  Apply compare_invert.

  Symmetry.
  Apply compare_invert.
Qed.

Lemma compare_equal_sym : (x, y : key)
  ((compare x y)=EQUAL) -> ((compare y x) = EQUAL).
Proof.
  Intros x y He.
  Replace (compare y x) with (invert_compare_result (compare x y)).
  Rewrite -> He.
  Simpl; Trivial.

  Symmetry.
  Apply compare_invert.
Qed.

Lemma invert_compare_result_invol : (x : compare_result)
  ((invert_compare_result (invert_compare_result x)) = x).
Proof.
  Intro.
  Unfold invert_compare_result.
  NewDestruct x; Trivial.
Qed.

Lemma compare_reverse1 : (x, y : key)
  ((compare x y)=GREATER) -> ((compare y x)=LESS).
Proof.
  Intros x y H.
  Replace LESS with (invert_compare_result (compare x y)).
  Apply compare_invert.

  Rewrite -> H.
  Unfold invert_compare_result.
  Trivial.
Qed.

Lemma compare_reverse2 : (x, y : key)
  ((compare x y)=LESS) -> ((compare y x)=GREATER).
Proof.
  Intros x y H.
  Replace GREATER with (invert_compare_result (compare x y)).
  Apply compare_invert.

  Rewrite -> H.
  Unfold invert_compare_result.
  Trivial.
Qed.

Lemma compare_refl : (x : key) ((compare x x) = EQUAL).
  Proof.
  Intro.
  Cut (cmp1:compare_result)(compare x x)=cmp1->(compare x x)=EQUAL.
  Intro W; Apply W with cmp1:=(compare x x); Trivial.

  Intros cmp1 HO1.
  NewDestruct cmp1.
  Cut (compare x x)=GREATER.
  Rewrite -> HO1; Intro Z; Discriminate Z.

  Apply compare_reverse2; Assumption.

  Assumption.

  Cut (compare x x)=LESS.
  Rewrite -> HO1; Intro Z; Discriminate Z.

  Apply compare_reverse1; Assumption.
Qed.

Lemma compare_antisym : (x, y : key)
  ((compare x y) = LESS) -> ((compare y x) = LESS) -> False.
Proof.
  Intros x y H1 H2.
  Cut (compare x y)=GREATER.
  Rewrite -> H1.
  Intro absurd; Discriminate absurd.

  Apply compare_reverse2.
  Assumption.
Qed.

Lemma compare_antisym2 : (x, y : key)
  ((compare x y) = GREATER) -> ((compare y x) = GREATER) -> False.
Proof.
  Intros x y H1 H2.
  Cut (compare x y)=LESS.
  Rewrite -> H1.
  Intro absurd; Discriminate absurd.

  Apply compare_reverse1.
  Assumption.
Qed.
