Inductive compare_result : Set :=
  LESS : compare_result
| EQUAL : compare_result
| GREATER : compare_result.

Definition invert_compare_result := [x : compare_result]
  Cases x of
  | LESS => GREATER
  | EQUAL => EQUAL
  | GREATER => LESS
  end.

(* 
Module Type ORDERED_SET.
  Parameter t: Set.
  Parameter compare: t->t->compare_result.
  Parameter compare_invert : (x, y : t)
    ((compare x y) = (invert_compare_result (compare y x))).
  Parameter compare_transitive : (x, y, z : t)
    ((compare x y) = LESS) ->  ((compare y z) = LESS) ->
    ((compare x z) = LESS).
  Parameter compare_equiv : (x, y, z : t)
    ((compare x y) = EQUAL) -> ((compare x z) = (compare y z)).
End ORDERED_SET.

Module Make [codomain : ORDERED_SET].
  Definition key := codomain.t.
  Definition compare : (key -> key -> compare_result) := codomain.compare.
  Definition compare_trans1 : (x, y, z : key)
     ((compare x y) = LESS) ->  ((compare y z) = LESS) -> ((compare x z) = LESS)
     := codomain.compare_transitive.
  Definition compare_equiv : (x, y, z : key)
    ((compare x y) = EQUAL) -> ((compare x z) = (compare y z))
    := codomain.compare_equiv.
  Definition compare_invert :
    (x, y : key) ((compare x y) = (invert_compare_result (compare y x)))
    := codomain.compare_invert.
*)

Parameter key : Set.
Parameter compare : (key -> key -> compare_result).
Parameter compare_trans1 : (x, y, z : key)
   ((compare x y) = LESS) ->  ((compare y z) = LESS) ->
   ((compare x z) = LESS).
Parameter compare_equiv : (x, y, z : key)
  ((compare x y) = EQUAL) -> ((compare x z) = (compare y z)).
Parameter compare_invert :
 (x, y : key) ((compare x y) = (invert_compare_result (compare y x))).

Load map_compare.

Section image_set.
  Variable a : Set.

  Inductive t : Set :=
  | Empty : t
  | Node : (left : t) (k : key) (data : a) (right : t) (height : nat) t.

  Definition height [tree : t] :=
    Cases tree of
    | Empty => O
    | (Node _ _ _ _ h) => h
    end.

  Require Max.
  Require Compare_dec.

  Load map_wf_order.
  Load map_create.
  Load map_find.
  Load map_bal.
  Load map_add.
  Load map_min_binding.
  Load map_merge.
  Load map_remove.
  Load map_rotate.
End image_set.

Load map_map.
Load map_map2.

(* End Make.*)

(* balancing
Require Max.

Inductive wf_height : t->Prop :=
| empty_wf_height : (wf_height Empty)
| node_wf_height : (left : t) (k : key) (data : a) (right : t) (wf_height left) -> (wf_height right) -> (wf_height (Node left k data right (S (max (height left) (height right))))).

Lemma create_wf_height : (left : t) (left_wf : (wf_height left))
(k : key) (data : a) (right : t) (right_wf : (wf_height right))
(wf_height (create left k data right)).
Proof.
Intros.
Unfold create.
Apply node_wf_height.
Assumption.

Assumption.
Qed.

Lemma bal_wf_height : (left : t) (left_wf : (wf_height left))
(k : key) (data : a) (right : t) (right_wf : (wf_height right))
(wf_height (bal left k data right)).
*)
