(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Fixpoint coqfind [tree : t] : key -> (option a) := [k : key]
  Cases tree of
  | Empty => (None a)
  | (Node l v d r _) =>
     Cases (compare k v) of
     | EQUAL => (Some a d)
     | LESS => (coqfind l k)
     | RIGHT => (coqfind r k)
     end
   end.

Theorem coqfind_equiv : (tree : t) (k1, k2 : key)
  (equiv : (compare k1 k2)= EQUAL) (coqfind tree k1)=(coqfind tree k2).
Proof.
  Intro tree.
  Elim tree.
  Simpl.
  Auto.

  Intros l Hl k d r Hr h k1 k2 equiv.
  Simpl.
  Replace (compare k2 k) with (compare k1 k).
  NewDestruct (compare k1 k).
  Apply Hl.
  Assumption.

  Trivial.

  Apply Hr.
  Assumption.

  Apply compare_equiv.
  Assumption.
Qed.

Lemma no_find_over : (tree : t) (k : key) (Hbound : (bounded_up k tree))
  ((coqfind tree k) = (None a)).
Proof.
  Intro.
  Elim tree.
  Simpl.
  Intros; Trivial.

  Intros l Hl k data r Hr h Klook Hbound.
  Inversion Hbound.
  Simpl.
  Cut (compare Klook k)=GREATER.
  Intro rew.
  Rewrite -> rew.
  Apply Hr.
  Assumption.

  Apply compare_reverse2.
  Assumption.
Qed.

Lemma no_find_under : (tree : t) (k : key) (Hbound : (bounded_down k tree))
  ((coqfind tree k) = (None a)).
Proof.
  Intro.
  Elim tree.
  Simpl.
  Intros; Trivial.

  Intros l Hl k data r Hr h Klook Hbound.
  Inversion Hbound.
  Simpl.
  Cut (compare Klook k)=LESS.
  Intro rew.
  Rewrite -> rew.
  Apply Hl.
  Assumption.

  Apply compare_reverse1.
  Assumption.
Qed.

Definition coqmem := [tree : t] [k : key]
  Cases (coqfind tree k) of
    None => false
  | (Some _) => true
  end.

