(* let rec map2iz f m1 m2 =
      if m1==m2 then m1 else
      match m1 with
        Empty -> if m2=Empty then Empty
        else raise (Invalid_argument "Map2.map2iz")
      | Node (l1,v1,d1,r1,h1) ->
          let (l2,d2,r2) = try cut v1 m2
          with Not_found -> raise (Invalid_argument "Map2.map2iz")
          in
          let dd = if d1==d2 then d1 else f v1 d1 d2 in
          Node (map2iz f l1 l2,v1,dd,map2iz f r1 r2,h1) *)

Section map2inz_f.
  Variable a: Set.
  Variable b: Set.
  Variable c: Set.
  Variable f: key->a->b->c.

  Fixpoint map2inz [tree1 : (t a)] : (t b) -> (t c) := [tree2 : (t b)]
    Cases tree1 of
    | Empty => (Empty c)
    | (Node l1 v1 d1 r1 h1) =>
        Cases (rotate b tree2 v1) of
        | Empty => (Empty c)
        | (Node l2 v2 d2 r2 _) =>
          (Node c (map2inz l1 l2) v1 (f v1 d1 d2) (map2inz r1 r2) h1)
        end
    end.

  Lemma map2inz_bounded_down : (tree1 : (t a)) (tree2 : (t b))
    (y : key)
    (bounded_down a y tree1) ->  (bounded_down b y tree2) ->
    (bounded_down c y (map2inz tree1 tree2)).
  Proof.
    Intro.
    Elim tree1.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion H1.
    NewDestruct tree2.
    Simpl.
    Constructor.

    Cut
     (rotated:(t b))
       (rotate b (Node b left1 k1 data1 right1 height2) k)=rotated
       ->  (bounded_down c y
     Cases (rotate b (Node b left1 k1 data1 right1 height2) k) of
       Empty => (Empty c)
     | (Node l2 _ d2 r2 _) => 
        (Node c (map2inz left l2) k (f k data d2) (map2inz right r2)
          height0)
     end).
    Intro W;
     Apply
      W with rotated:=(rotate b (Node b left1 k1 data1 right1 height2) k);
     Trivial.

    Intros rotated Rrotated.
    Rewrite -> Rrotated.
    NewDestruct rotated.
    Constructor.

    Cut
     (bounded_down b y (rotate b (Node b left1 k1 data1 right1 height2) k)).
    Rewrite -> Rrotated.
    Intro Hbounded.
    Inversion Hbounded.
    Constructor.
    Apply H; Assumption.

    Replace GREATER with (compare k2 y).
    Apply compare_equiv.
    Cut
     Cases (rotate b (Node b left1 k1 data1 right1 height2) k) of
       Empty => True
     | (Node _ w _ _ _) => (compare k w)=EQUAL
     end.
    Rewrite -> Rrotated.
    Tauto.

    Apply rotate_returns_center.

    Apply H0; Assumption.

    Apply rotate_bounded_down; Assumption.
  Qed.

  Lemma map2inz_bounded_up : (tree1 : (t a)) (tree2 : (t b))
    (y : key)
    (bounded_up a y tree1) ->  (bounded_up b y tree2) ->
    (bounded_up c y (map2inz tree1 tree2)).
  Proof.
    Intro.
    Elim tree1.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion H1.
    NewDestruct tree2.
    Simpl.
    Constructor.

    Cut
     (rotated:(t b))
       (rotate b (Node b left1 k1 data1 right1 height2) k)=rotated
       ->  (bounded_up c y
     Cases (rotate b (Node b left1 k1 data1 right1 height2) k) of
       Empty => (Empty c)
     | (Node l2 _ d2 r2 _) => 
        (Node c (map2inz left l2) k (f k data d2) (map2inz right r2)
          height0)
     end).
    Intro W;
     Apply
      W with rotated:=(rotate b (Node b left1 k1 data1 right1 height2) k);
     Trivial.

    Intros rotated Rrotated.
    Rewrite -> Rrotated.
    NewDestruct rotated.
    Constructor.

    Cut
     (bounded_up b y (rotate b (Node b left1 k1 data1 right1 height2) k)).
    Rewrite -> Rrotated.
    Intro Hbounded.
    Inversion Hbounded.
    Constructor.
    Apply H; Assumption.

    Replace LESS with (compare k2 y).
    Apply compare_equiv.
    Cut
     Cases (rotate b (Node b left1 k1 data1 right1 height2) k) of
       Empty => True
     | (Node _ w _ _ _) => (compare k w)=EQUAL
     end.
    Rewrite -> Rrotated.
    Tauto.

    Apply rotate_returns_center.

    Apply H0; Assumption.

    Apply rotate_bounded_up; Assumption.
  Qed.

  Lemma map2inz_wf_order : (tree1 : (t a)) (tree2 : (t b))
    (wf_order a tree1) ->  (wf_order b tree2) ->
    (wf_order c (map2inz tree1 tree2)).
  Proof.
    Intro.
    Elim tree1.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion H1.
    NewDestruct tree2.
    Simpl.
    Constructor.

    Cut
     (rotated:(t b))
       (rotate b (Node b left1 k1 data1 right1 height2) k)=rotated
       ->(wf_order c
           Cases (rotate b (Node b left1 k1 data1 right1 height2) k) of
       Empty => (Empty c)
     | (Node l2 _ d2 r2 _) => 
        (Node c (map2inz left l2) k (f k data d2) (map2inz right r2)
          height0)
     end).
    Intro W;
     Apply
      W with rotated:=(rotate b (Node b left1 k1 data1 right1 height2) k);
     Trivial.

    Intros rotated Rrotated.
    Rewrite -> Rrotated.
    NewDestruct rotated.
    Constructor.

    Cut (wf_order b (rotate b (Node b left1 k1 data1 right1 height2) k)).
    Rewrite -> Rrotated.
    Intro Hwf_order.
    Inversion Hwf_order.
    Cut (compare k k2)=EQUAL.
    Intro equal.
    Constructor.
    Apply H; Assumption.

    Apply H0; Assumption.

    Apply map2inz_bounded_up.
    Assumption.

    Apply wf_order_overbound_equal with k:=k2; Assumption.

    Apply map2inz_bounded_down.
    Assumption.

    Apply wf_order_underbound_equal with k:=k2; Assumption.

    Cut
     Cases (rotate b (Node b left1 k1 data1 right1 height2) k) of
       Empty => True
     | (Node _ w _ _ _) => (compare k w)=EQUAL
     end.
    Rewrite -> Rrotated.
    Tauto.

    Apply rotate_returns_center.

    Apply rotate_wf_order; Assumption.
  Qed.

  Definition func_map2inz :=
    [m1 : key->(option a)] [m2 : key -> (option b)] [k : key] 
    Cases (m1 k) of
    | None => (None c)
    | (Some v1) =>
       Cases (m2 k) of
       | None => (None c)
       | (Some v2) => (Some c (f k v1 v2))
       end
    end.

  Variable f_respects: (k1, k2: key) (compare k1 k2)=EQUAL ->
    (va : a) (vb : b) (f k1 va vb) = (f k2 va vb).

  Lemma rotate_cannot_empty :
    (l1 : (t a)) (v1 : key) (d1 : a) (r1 : (t a)) (h1 : nat)
    (tree2 : (t b))
    (same_domain : (xcon : key)
       (coqmem a (Node a l1 v1 d1 r1 h1) xcon)=(coqmem b tree2 xcon))
    (rotate b tree2 v1) = (Empty b) -> False.
  Proof.
    Intros.
    Cut
     Cases (coqfind b tree2 v1) of
       None => True
     | (Some d) => (rotate b tree2 v1)=(Empty b)->False
     end.
    Cut (coqmem a (Node a l1 v1 d1 r1 h1) v1)=(coqmem b tree2 v1).
    Unfold coqmem.
    Simpl.
    Rewrite -> compare_refl.
    NewDestruct (coqfind b tree2 v1).
    Rewrite -> H.
    Auto.

    Intro Z; Discriminate Z.

    Apply same_domain.

    Apply rotate_not_empty.
  Qed.

  Lemma rotate_preserves_coqmem :
     (tree : (t b)) (v : key)
     (tree_wf_order : (wf_order b tree))
     (coqmem b tree v)=true -> (k_find : key)
     (coqmem b (rotate b tree v) k_find)=(coqmem b tree k_find).
  Proof.
    Unfold coqmem.
    Intros.
    Cut
     Cases (coqfind b tree v) of
       None => True
     | (Some _) =>
        (coqfind b (rotate b tree v) k_find)=(coqfind b tree k_find)
     end.
    NewDestruct (coqfind b tree v).
    Intro rew; Rewrite -> rew.
    Trivial.

    Discriminate H.

    Apply rotate_correct.
    Assumption.
  Qed.

  Theorem map2inz_correct :
    (tree1 : (t a)) (tree2 : (t b))
    (tree1_wf_order : (wf_order a tree1)) (tree2_wf_order : (wf_order b tree2))
    (same_domain : (xcon : key) (coqmem a tree1 xcon)=(coqmem b tree2 xcon))
    (k : key)
    (coqfind c (map2inz tree1 tree2) k) =
    (func_map2inz (coqfind a tree1) (coqfind b tree2) k).
  Proof.
    Intro tree1.
    Elim tree1.
    Unfold func_map2inz.
    Simpl.
    Auto.

    Unfold func_map2inz.
    Intros l1 Hl1 v1 d1 r1 Hr1 height0 tree2 tree1_wf_order tree2_wf_order
     same_domain k.
    Simpl.
    Cut
     (rtree2:(t b))
       (rotate b tree2 v1)=rtree2
       ->(coqfind c
           Cases (rotate b tree2 v1) of
             Empty => (Empty c)
           | (Node l2 _ d2 r2 _) =>
              (Node c (map2inz l1 l2) v1 (f v1 d1 d2) (map2inz r1 r2) height0)
           end k)
          =Cases
             Cases (compare k v1) of
               LESS => (coqfind a l1 k)
             | EQUAL => (Some a d1)
             | GREATER => (coqfind a r1 k)
             end
           of
             (Some v0) =>
              Cases (coqfind b tree2 k) of
                (Some v2) => (Some c (f k v0 v2))
              | None => (None c)
              end
           | None => (None c)
           end.
    Intro W; Apply W with rtree2:=(rotate b tree2 v1); Trivial.

    Intros rtree2 Hrtree2.
    Rewrite -> Hrtree2.
    NewDestruct rtree2.
    Simpl.
    Apply False_ind.
    Apply
     rotate_cannot_empty
                         with
                         l1:=l1
                        v1:=v1
                        d1:=d1
                        r1:=r1
                        h1:=height0
                        tree2:=tree2.
    Assumption.

    Assumption.

    Simpl.
    Cut (coqfind b tree2 k)=(coqfind b (rotate b tree2 v1) k).
    Intro equivb.
    Rewrite -> equivb.
    Rewrite -> Hrtree2.
    Simpl.
    Replace (compare k k0) with (compare k v1).
    TCompare k v1 Hkv1.
    Apply Hl1.
    Inversion tree1_wf_order; Assumption.

    Cut (wf_order b (rotate b tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Intro xcon.
    TCompare xcon v1 Hxconv1.
    Replace (coqmem a l1 xcon) with
    (coqmem a (Node a l1 v1 d1 r1 height0) xcon).
    Replace (coqmem b left xcon) with (coqmem b tree2 xcon).
    Apply same_domain.

    Replace (coqmem b tree2 xcon) with (coqmem b (rotate b tree2 v1) xcon).
    Rewrite -> Hrtree2.
    Simpl.
    Unfold coqmem.
    Simpl.
    Replace (compare xcon k0) with (compare xcon v1).
    Rewrite -> Hxconv1.
    Trivial.

    Apply compare_equiv2.
    Cut
     Cases (rotate b tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Trivial.

    Apply rotate_returns_center.

    Apply rotate_preserves_coqmem.
    Assumption.

    Rewrite <- same_domain.
    Unfold coqmem.
    Simpl.
    Rewrite -> compare_refl.
    Trivial.

    Unfold coqmem.
    Simpl.
    Rewrite -> Hxconv1.
    Trivial.

    Unfold coqmem.
    Rewrite -> no_find_over.
    Rewrite -> no_find_over.
    Trivial.

    Apply wf_order_overbound_equal with k:=k0.
    Replace EQUAL with (compare xcon v1).
    Apply compare_equiv2.
    Cut
     Cases (rotate b tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Intro Z.
    Apply compare_equal_sym; Assumption.

    Apply rotate_returns_center.

    Cut (wf_order b (rotate b tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Apply wf_order_overbound_equal with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Unfold coqmem.
    Rewrite -> no_find_over.
    Rewrite -> no_find_over.
    Trivial.

    Apply wf_order_overbound with k:=v1.
    Assumption.

    Apply wf_order_overbound_equal with k:=k0.
    Cut
     Cases (rotate b tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Trivial.

    Apply rotate_returns_center.

    Cut (wf_order b (rotate b tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Apply wf_order_overbound with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Rewrite -> f_respects with k1:=v1 k2:=k va:=d1 vb:=data.
    Trivial.

    Apply compare_equal_sym; Assumption.

    Cut (compare v1 k0)=EQUAL.
    Intro equal.
    Apply Hr1.
    Inversion tree1_wf_order; Assumption.

    Cut (wf_order b (rotate b tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Intro xcon.
    TCompare xcon v1 Hxconv1.
    Unfold coqmem.
    Rewrite -> no_find_under.
    Rewrite -> no_find_under.
    Trivial.

    Apply wf_order_underbound with k:=v1.
    Assumption.

    Apply wf_order_underbound_equal with k:=k0.
    Assumption.

    Cut (wf_order b (rotate b tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order; Assumption.

    Apply wf_order_underbound with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Unfold coqmem.
    Rewrite -> no_find_under.
    Rewrite -> no_find_under.
    Trivial.

    Apply wf_order_underbound_equal with k:=v1.
    Assumption.

    Apply wf_order_underbound_equal with k:=k0.
    Assumption.

    Cut (wf_order b (rotate b tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Apply wf_order_underbound_equal with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Replace (coqmem a r1 xcon) with
    (coqmem a (Node a l1 v1 d1 r1 height0) xcon).
    Replace (coqmem b right xcon) with (coqmem b (rotate b tree2 v1) xcon).
    Rewrite -> rotate_preserves_coqmem.
    Apply same_domain.

    Assumption.

    Rewrite <- same_domain.
    Unfold coqmem; Simpl; Rewrite -> compare_refl; Trivial.

    Rewrite -> Hrtree2.
    Unfold coqmem.
    Simpl.
    Replace (compare xcon k0) with GREATER.
    Trivial.

    Rewrite <- Hxconv1.
    Apply compare_equiv2.
    Assumption.

    Unfold coqmem.
    Simpl.
    Rewrite -> Hxconv1.
    Trivial.

    Cut
     Cases (rotate b tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Auto.

    Apply rotate_returns_center.

    Apply compare_equiv2.
    Cut
     Cases (rotate b tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Auto.

    Apply rotate_returns_center.

    Cut
     Cases (coqfind b tree2 v1) of
       None => True
     | (Some _) => (coqfind b (rotate b tree2 v1) k)=(coqfind b tree2 k)
     end.
    Cut (coqmem a (Node a l1 v1 d1 r1 height0) v1)=(coqmem b tree2 v1).
    Unfold coqmem.
    Simpl.
    Rewrite -> compare_refl.
    NewDestruct (coqfind b tree2 v1).
    Auto.

    Intro Z; Discriminate Z.

    Apply same_domain.

    Apply rotate_correct.
    Assumption.
  Qed.
End map2inz_f.

Parameter physical_equal : (S : Set) (x : S) (y : S) { x = y } + { True }.

Section map2iz_f.
  Variable a: Set.
  Variable f: key->a->a->a.
  Variable f_respects : (k1, k2: key) (x, y : a) (f k1 x y)=(f k2 x y).

  Fixpoint map2iz [tree1 : (t a)] : (t a) -> (t a) := [tree2 : (t a)]
    Cases (physical_equal (t a) tree1 tree2) of
    | (left _) => tree1
    | (right _) =>
      Cases tree1 of
      | Empty => (Empty a)
      | (Node l1 v1 d1 r1 h1) =>
          Cases (rotate a tree2 v1) of
          | Empty => (Empty a)
          | (Node l2 v2 d2 r2 _) =>
            (Node a (map2iz l1 l2) v1 (f v1 d1 d2) (map2iz r1 r2) h1)
          end
      end
    end.

  Lemma map2iz_bounded_down : (tree1 : (t a)) (tree2 : (t a))
    (y : key)
    (bounded_down a y tree1) ->  (bounded_down a y tree2) ->
    (bounded_down a y (map2iz tree1 tree2)).
  Proof.
    Intro.
    Elim tree1.
    Simpl.
    Intros.
    NewDestruct (physical_equal (t a) (Empty a) tree2); Constructor.

    Intros.
    Simpl.
    NewDestruct
     (physical_equal (t a) (Node a left k data right height0) tree2).
    Assumption.

    Clear t0.
    Inversion H1.
    NewDestruct tree2.
    Simpl.
    Constructor.

    Cut
     (rotated:(t a))
       (rotate a (Node a left1 k1 data1 right1 height2) k)=rotated
       ->(bounded_down a y
           Cases (rotate a (Node a left1 k1 data1 right1 height2) k) of
             Empty => (Empty a)
           | (Node l2 _ d2 r2 _) =>
              (Node a (map2iz left l2) k (f k data d2) (map2iz right r2)
                height0)
           end).
    Intro W;
     Apply
      W with rotated:=(rotate a (Node a left1 k1 data1 right1 height2) k);
     Trivial.

    Intros rotated Rrotated.
    Rewrite -> Rrotated.
    NewDestruct rotated.
    Constructor.

    Cut
     (bounded_down a y (rotate a (Node a left1 k1 data1 right1 height2) k)).
    Rewrite -> Rrotated.
    Intro Hbounded.
    Inversion Hbounded.
    Constructor.
    Apply H; Assumption.

    Replace GREATER with (compare k2 y).
    Apply compare_equiv.
    Cut
     Cases (rotate a (Node a left1 k1 data1 right1 height2) k) of
       Empty => True
     | (Node _ w _ _ _) => (compare k w)=EQUAL
     end.
    Rewrite -> Rrotated.
    Tauto.

    Apply rotate_returns_center.

    Apply H0; Assumption.

    Apply rotate_bounded_down; Assumption.
  Qed.

  Lemma map2iz_bounded_up : (tree1 : (t a)) (tree2 : (t a))
    (y : key)
    (bounded_up a y tree1) ->  (bounded_up a y tree2) ->
    (bounded_up a y (map2iz tree1 tree2)).
  Proof.
    Intro.
    Elim tree1.
    Simpl.
    Intros.
    NewDestruct (physical_equal (t a) (Empty a) tree2); Constructor.

    Intros.
    Simpl.
    NewDestruct
     (physical_equal (t a) (Node a left k data right height0) tree2).
    Assumption.

    Clear t0.
    Inversion H1.
    NewDestruct tree2.
    Simpl.
    Constructor.

    Cut
     (rotated:(t a))
       (rotate a (Node a left1 k1 data1 right1 height2) k)=rotated
       ->(bounded_up a y
           Cases (rotate a (Node a left1 k1 data1 right1 height2) k) of
             Empty => (Empty a)
           | (Node l2 _ d2 r2 _) =>
              (Node a (map2iz left l2) k (f k data d2) (map2iz right r2)
                height0)
           end).
    Intro W;
     Apply
      W with rotated:=(rotate a (Node a left1 k1 data1 right1 height2) k);
     Trivial.

    Intros rotated Rrotated.
    Rewrite -> Rrotated.
    NewDestruct rotated.
    Constructor.

    Cut
     (bounded_up a y (rotate a (Node a left1 k1 data1 right1 height2) k)).
    Rewrite -> Rrotated.
    Intro Hbounded.
    Inversion Hbounded.
    Constructor.
    Apply H; Assumption.

    Replace LESS with (compare k2 y).
    Apply compare_equiv.
    Cut
     Cases (rotate a (Node a left1 k1 data1 right1 height2) k) of
       Empty => True
     | (Node _ w _ _ _) => (compare k w)=EQUAL
     end.
    Rewrite -> Rrotated.
    Tauto.

    Apply rotate_returns_center.

    Apply H0; Assumption.

    Apply rotate_bounded_up; Assumption.
  Qed.

  Lemma map2iz_wf_order : (tree1 : (t a)) (tree2 : (t a))
    (wf_order a tree1) ->  (wf_order a tree2) ->
    (wf_order a (map2iz tree1 tree2)).
  Proof.
    Intro.
    Elim tree1.
    Simpl.
    Intros.
    NewDestruct (physical_equal (t a) (Empty a) tree2); Constructor.

    Intros.
    Simpl.
    NewDestruct
     (physical_equal (t a) (Node a left k data right height0) tree2).
    Assumption.

    Clear t0.
    Inversion H1.
    NewDestruct tree2.
    Simpl.
    Constructor.

    Cut
     (rotated:(t a))
       (rotate a (Node a left1 k1 data1 right1 height2) k)=rotated
       ->(wf_order a
           Cases (rotate a (Node a left1 k1 data1 right1 height2) k) of
       Empty => (Empty a)
     | (Node l2 _ d2 r2 _) => 
        (Node a (map2iz left l2) k (f k data d2) (map2iz right r2)
          height0)
     end).
    Intro W;
     Apply
      W with rotated:=(rotate a (Node a left1 k1 data1 right1 height2) k);
     Trivial.

    Intros rotated Rrotated.
    Rewrite -> Rrotated.
    NewDestruct rotated.
    Constructor.

    Cut (wf_order a (rotate a (Node a left1 k1 data1 right1 height2) k)).
    Rewrite -> Rrotated.
    Intro Hwf_order.
    Inversion Hwf_order.
    Cut (compare k k2)=EQUAL.
    Intro equal.
    Constructor.
    Apply H; Assumption.

    Apply H0; Assumption.

    Apply map2iz_bounded_up.
    Assumption.

    Apply wf_order_overbound_equal with k:=k2; Assumption.

    Apply map2iz_bounded_down.
    Assumption.

    Apply wf_order_underbound_equal with k:=k2; Assumption.

    Cut
     Cases (rotate a (Node a left1 k1 data1 right1 height2) k) of
       Empty => True
     | (Node _ w _ _ _) => (compare k w)=EQUAL
     end.
    Rewrite -> Rrotated.
    Tauto.

    Apply rotate_returns_center.

    Apply rotate_wf_order; Assumption.
  Qed.

  Variable f_idempotent : (k : key) (x : a) (f k x x)=x.
  
  Theorem map2iz_correct :
    (tree1, tree2 : (t a))
    (tree1_wf_order : (wf_order a tree1)) (tree2_wf_order : (wf_order a tree2))
    (same_domain : (xcon : key) (coqmem a tree1 xcon)=(coqmem a tree2 xcon))
    (k : key)
    (coqfind a (map2iz tree1 tree2) k) =
    (func_map2inz a a a f (coqfind a tree1) (coqfind a tree2) k).
  Proof.
    Intro tree1.
    Elim tree1.
    Unfold func_map2inz.
    Simpl.
    Intros.
    NewDestruct (physical_equal (t a) (Empty a) tree2); Simpl; Auto.

    Unfold func_map2inz.
    Intros l1 Hl1 v1 d1 r1 Hr1 height0 tree2 tree1_wf_order tree2_wf_order
     same_domain k.
    Simpl.
    NewDestruct (physical_equal (t a) (Node a l1 v1 d1 r1 height0) tree2).
    Rewrite <- e.
    Simpl.
    TCompare k v1 Tkv1.
    NewDestruct (coqfind a l1 k).
    Rewrite -> f_idempotent; Trivial.

    Trivial.

    Rewrite -> f_idempotent; Trivial.

    NewDestruct (coqfind a r1 k).
    Rewrite -> f_idempotent; Trivial.

    Trivial.

    Clear t0.
    Cut
     (rtree2:(t a))
       (rotate a tree2 v1)=rtree2
       ->(coqfind a
           Cases (rotate a tree2 v1) of
             Empty => (Empty a)
           | (Node l2 _ d2 r2 _) =>
              (Node a (map2iz l1 l2) v1 (f v1 d1 d2) (map2iz r1 r2) height0)
           end k)
          =Cases
             Cases (compare k v1) of
               LESS => (coqfind a l1 k)
             | EQUAL => (Some a d1)
             | GREATER => (coqfind a r1 k)
             end
           of
             (Some v0) =>
              Cases (coqfind a tree2 k) of
                (Some v2) => (Some a (f k v0 v2))
              | None => (None a)
              end
           | None => (None a)
           end.
    Intro W; Apply W with rtree2:=(rotate a tree2 v1); Trivial.

    Intros rtree2 Hrtree2.
    Rewrite -> Hrtree2.
    NewDestruct rtree2.
    Simpl.
    Apply False_ind.
    Apply
     rotate_cannot_empty
                         with
                         l1:=l1
                        v1:=v1
                        d1:=d1
                        r1:=r1
                        h1:=height0
                        tree2:=tree2.
    Assumption.

    Assumption.

    Simpl.
    Cut (coqfind a tree2 k)=(coqfind a (rotate a tree2 v1) k).
    Intro equivb.
    Rewrite -> equivb.
    Rewrite -> Hrtree2.
    Simpl.
    Replace (compare k k0) with (compare k v1).
    TCompare k v1 Hkv1.
    Apply Hl1.
    Inversion tree1_wf_order; Assumption.

    Cut (wf_order a (rotate a tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Intro xcon.
    TCompare xcon v1 Hxconv1.
    Replace (coqmem a l1 xcon) with
    (coqmem a (Node a l1 v1 d1 r1 height0) xcon).
    Replace (coqmem a left xcon) with (coqmem a tree2 xcon).
    Apply same_domain.

    Replace (coqmem a tree2 xcon) with (coqmem a (rotate a tree2 v1) xcon).
    Rewrite -> Hrtree2.
    Simpl.
    Unfold coqmem.
    Simpl.
    Replace (compare xcon k0) with (compare xcon v1).
    Rewrite -> Hxconv1.
    Trivial.

    Apply compare_equiv2.
    Cut
     Cases (rotate a tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Trivial.

    Apply rotate_returns_center.

    Apply rotate_preserves_coqmem.
    Assumption.

    Rewrite <- same_domain.
    Unfold coqmem.
    Simpl.
    Rewrite -> compare_refl.
    Trivial.

    Unfold coqmem.
    Simpl.
    Rewrite -> Hxconv1.
    Trivial.

    Unfold coqmem.
    Rewrite -> no_find_over.
    Rewrite -> no_find_over.
    Trivial.

    Apply wf_order_overbound_equal with k:=k0.
    Replace EQUAL with (compare xcon v1).
    Apply compare_equiv2.
    Cut
     Cases (rotate a tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Intro Z.
    Apply compare_equal_sym; Assumption.

    Apply rotate_returns_center.

    Cut (wf_order a (rotate a tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Apply wf_order_overbound_equal with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Unfold coqmem.
    Rewrite -> no_find_over.
    Rewrite -> no_find_over.
    Trivial.

    Apply wf_order_overbound with k:=v1.
    Assumption.

    Apply wf_order_overbound_equal with k:=k0.
    Cut
     Cases (rotate a tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Trivial.

    Apply rotate_returns_center.

    Cut (wf_order a (rotate a tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Apply wf_order_overbound with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Rewrite -> f_respects with k1:=v1 k2:=k x:=d1 y:=data.
    Trivial.

    Cut (compare v1 k0)=EQUAL.
    Intro equal.
    Apply Hr1.
    Inversion tree1_wf_order; Assumption.

    Cut (wf_order a (rotate a tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Intro xcon.
    TCompare xcon v1 Hxconv1.
    Unfold coqmem.
    Rewrite -> no_find_under.
    Rewrite -> no_find_under.
    Trivial.

    Apply wf_order_underbound with k:=v1.
    Assumption.

    Apply wf_order_underbound_equal with k:=k0.
    Assumption.

    Cut (wf_order a (rotate a tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order; Assumption.

    Apply wf_order_underbound with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Unfold coqmem.
    Rewrite -> no_find_under.
    Rewrite -> no_find_under.
    Trivial.

    Apply wf_order_underbound_equal with k:=v1.
    Assumption.

    Apply wf_order_underbound_equal with k:=k0.
    Assumption.

    Cut (wf_order a (rotate a tree2 v1)).
    Rewrite -> Hrtree2.
    Intro Z; Inversion Z; Assumption.

    Apply rotate_wf_order.
    Assumption.

    Apply wf_order_underbound_equal with k:=v1.
    Assumption.

    Inversion tree1_wf_order; Assumption.

    Replace (coqmem a r1 xcon) with
    (coqmem a (Node a l1 v1 d1 r1 height0) xcon).
    Replace (coqmem a right xcon) with (coqmem a (rotate a tree2 v1) xcon).
    Rewrite -> rotate_preserves_coqmem.
    Apply same_domain.

    Assumption.

    Rewrite <- same_domain.
    Unfold coqmem; Simpl; Rewrite -> compare_refl; Trivial.

    Rewrite -> Hrtree2.
    Unfold coqmem.
    Simpl.
    Replace (compare xcon k0) with GREATER.
    Trivial.

    Rewrite <- Hxconv1.
    Apply compare_equiv2.
    Assumption.

    Unfold coqmem.
    Simpl.
    Rewrite -> Hxconv1.
    Trivial.

    Cut
     Cases (rotate a tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Auto.

    Apply rotate_returns_center.

    Apply compare_equiv2.
    Cut
     Cases (rotate a tree2 v1) of
       Empty => True
     | (Node _ k _ _ _) => (compare v1 k)=EQUAL
     end.
    Rewrite -> Hrtree2.
    Auto.

    Apply rotate_returns_center.

    Cut
     Cases (coqfind a tree2 v1) of
       None => True
     | (Some _) => (coqfind a (rotate a tree2 v1) k)=(coqfind a tree2 k)
     end.
    Cut (coqmem a (Node a l1 v1 d1 r1 height0) v1)=(coqmem a tree2 v1).
    Unfold coqmem.
    Simpl.
    Rewrite -> compare_refl.
    NewDestruct (coqfind a tree2 v1).
    Auto.

    Intro Z; Discriminate Z.

    Apply same_domain.

    Apply rotate_correct.
    Assumption.
  Qed.
End map2iz_f.
