(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Definition merge := [t1 : t] [t2 : t]
  Cases t1 of
  | Empty => t2
  | (Node l1 x1 d1 r1 _) =>
    Cases t2 of
    | Empty => t1
    | (Node l2 x2 d2 r2 _) =>
      let (x, d) = (min_binding l2 x2 d2) in
        (bal t1 x d (remove_min_binding l2 x2 d2 r2))
    end
  end.

Lemma merge_wf_order : (t1 : t) (t1_wf_order : (wf_order t1))
  (t2 : t) (t2_wf_order : (wf_order t2))
  (x : key) (t1_bounded_up : (bounded_up x t1))
  (t2_bounded_down : (bounded_down x t2))
  (wf_order (merge t1 t2)).
Proof.
  Intro t1.
  NewDestruct t1.
  Simpl.
  Auto.

  Intro t1_wf_order.
  Intro t2.
  NewDestruct t2.
  Simpl.
  Auto.

  Intro t2_wf_order.
  Intros.
  Simpl.
  Cut
   (x_min:key; d_min:a)
     (min_binding left0 k0 data0)=(pair key a x_min d_min)
     ->(wf_order
         let
          (x0,d) =
           (min_binding left0 k0 data0)
         in (bal (Node left k data right height0) x0 d
              (remove_min_binding left0 k0 data0 right0))).
  Intro Z.
  Apply
   Z
     with
     x_min:=(fst key a (min_binding left0 k0 data0))
    d_min:=(snd key a (min_binding left0 k0 data0)).
  NewDestruct (min_binding left0 k0 data0).
  Simpl; Trivial.

  Intros x_min d_min rew.
  Rewrite -> rew.
  Cut x_min=(fst key a (min_binding left0 k0 data0)).
  Intro Hx_min.
  Apply bal_wf_order.
  Assumption.

  Apply remove_min_binding_wf_order.
  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Apply wf_order_overbound with k:=x.
  Rewrite -> Hx_min.
  Apply compare_reverse2.
  Apply min_binding_bounded_down.
  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion t2_bounded_down; Assumption.

  Inversion t2_bounded_down.
  Apply compare_reverse1; Assumption.

  Assumption.

  Rewrite -> Hx_min.
  Apply min_binding_bounds_down_removed.
  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Rewrite -> rew.
  Simpl.
  Trivial.
Qed.

Lemma merge_bounded_up : (t1 : t) (t1_wf_order : (wf_order t1))
  (t2 : t) (t2_wf_order : (wf_order t2))
  (x : key) (t1_bounded_up : (bounded_up x t1))
  (t2_bounded_down : (bounded_down x t2))
  (y : key)
  (x_bounded_up : (compare x y)=LESS)
  (t2_bounded_up : (bounded_up y t2))
  (bounded_up y (merge t1 t2)).
Proof.
  Intro t1.
  NewDestruct t1.
  Simpl.
  Auto.

  Intros t1_wf_order t2 t2_wf_order x t1_bounded_up t2_bounded_down y
   x_bounded_up t2_bounded_up.
  Simpl.
  NewDestruct t2.
  Constructor.
  Apply wf_order_overbound with k:=x.
  Apply compare_reverse2; Assumption.

  Inversion t1_bounded_up; Assumption.

  Apply compare_trans1 with y:=x.
  Inversion t1_bounded_up; Assumption.

  Assumption.

  Apply wf_order_overbound with k:=x.
  Apply compare_reverse2; Assumption.

  Inversion t1_bounded_up; Assumption.

  Cut
   (x0:key; d:a)
     (x0,d)=(min_binding left0 k0 data0)
     ->(bounded_up y
         let
          (x0,d) =
           (min_binding left0 k0 data0)
         in (bal (Node left k data right height0) x0 d
              (remove_min_binding left0 k0 data0 right0))).
  Intro W;
   Apply
    W
      with
      x0:=(fst key a (min_binding left0 k0 data0))
     d:=(snd key a (min_binding left0 k0 data0)); Simpl.
  NewDestruct (min_binding left0 k0 data0); Simpl; Trivial.

  Intros x0 d Hx0d.
  Rewrite <- Hx0d.
  Apply bal_bounded_up.
  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply compare_reverse1.
  Apply min_binding_bounded_up.
  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion t2_bounded_up.
  Apply compare_reverse2; Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Apply wf_order_overbound with k:=x.
  Apply compare_reverse2; Assumption.

  Assumption.

  Apply remove_min_binding_bounded_up.
  Inversion t2_bounded_up; Assumption.

  Inversion t2_bounded_up; Assumption.

  Inversion t2_bounded_up; Assumption.
Qed.

Lemma merge_bounded_down : (t1 : t) (t1_wf_order : (wf_order t1))
  (t2 : t) (t2_wf_order : (wf_order t2))
  (x : key) (t1_bounded_up : (bounded_up x t1))
  (t2_bounded_down : (bounded_down x t2))
  (y : key)
  (x_bounded_down : (compare x y)=GREATER)
  (t1_bounded_down : (bounded_down y t1))
  (bounded_down y (merge t1 t2)).
Proof.
  Intro t1.
  NewDestruct t1.
  Simpl.
  Intros.
  Apply wf_order_underbound with k:=x.
  Apply compare_reverse1; Assumption.

  Assumption.

  Intros t1_wf_order t2 t2_wf_order x l_bounded_up r_bounded_down y
   x_bounded_down l_bounded_down.
  Simpl.
  NewDestruct t2.
  Assumption.

  Cut
   (x0:key; d:a)
     (x0,d)=(min_binding left0 k0 data0)
     ->(bounded_down y
         let
          (x0,d) =
           (min_binding left0 k0 data0)
         in (bal (Node left k data right height0) x0 d
              (remove_min_binding left0 k0 data0 right0))).
  Intro W.
  Apply
   W
     with
     x0:=(fst key a (min_binding left0 k0 data0))
    d:=(snd key a (min_binding left0 k0 data0)).
  NewDestruct (min_binding left0 k0 data0); Simpl; Trivial.

  Intros x0 d Hx0d.
  Rewrite <- Hx0d; Simpl.
  Apply bal_bounded_down.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=x.
  Apply compare_reverse1; Assumption.

  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounded_down.
  Inversion t2_wf_order; Assumption.

  Inversion t2_wf_order; Assumption.

  Inversion r_bounded_down; Assumption.

  Inversion r_bounded_down.
  Apply compare_reverse1; Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Assumption.

  Apply remove_min_binding_bounded_down.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=x.
  Apply compare_reverse1; Assumption.

  Inversion r_bounded_down.
  Apply compare_reverse1; Assumption.

  Apply wf_order_underbound with k:=x.
  Apply compare_reverse1; Assumption.

  Inversion r_bounded_down; Assumption.

  Apply wf_order_underbound with k:=x.
  Apply compare_reverse1; Assumption.

  Inversion r_bounded_down; Assumption.
Qed.

Lemma find_merge : (t1 : t)(t1_wf_order : (wf_order t1))
  (t2 : t) (t2_wf_order : (wf_order t2))
  (x : key) (t1_bounded_up : (bounded_up x t1))
  (t2_bounded_down : (bounded_down x t2))
  (k_find : key)
  (coqfind (merge t1 t2) k_find) =
  (Cases (compare k_find x) of
  | EQUAL => (None a)
  | LESS => (coqfind t1 k_find)
  | GREATER => (coqfind t2 k_find)
  end).
Proof.
  Intro t1.
  NewDestruct t1.
  Simpl.
  Simpl.
  Intros.
  Cut
   (cmp1:compare_result)
     (compare k_find x)=cmp1
     ->(coqfind t2 k_find)
        =Cases (compare k_find x) of
           LESS => (None a)
         | EQUAL => (None a)
         | GREATER => (coqfind t2 k_find)
         end.
  Intro Z.
  Apply Z with cmp1:=(compare k_find x); Trivial.

  Intros cmp1 HO1.
  NewDestruct cmp1.
  Rewrite -> HO1.
  Apply no_find_under.
  Apply wf_order_underbound with k:=x.
  Assumption.

  Assumption.

  Rewrite -> HO1.
  Apply no_find_under.
  Apply wf_order_underbound_equal with k:=x.
  Assumption.

  Assumption.

  Rewrite -> HO1.
  Trivial.

  Intros Hwf_order1 t2 Hwf_order2.
  NewDestruct t2.
  Intros.
  Cut
   (cmp1:compare_result)
     (compare k_find x)=cmp1
     ->(coqfind (merge (Node left k data right height0) Empty) k_find)
        =Cases (compare k_find x) of
           LESS => (coqfind (Node left k data right height0) k_find)
         | EQUAL => (None a)
         | GREATER => (coqfind Empty k_find)
         end.
  Intro Z.
  Apply Z with cmp1:=(compare k_find x); Trivial.

  Intros cmp1 HO1.
  NewDestruct cmp1.
  Rewrite -> HO1.
  Simpl.
  Trivial.

  Rewrite -> HO1.
  Simpl.
  Cut
   (cmp2:compare_result)
     (compare k_find k)=cmp2
     ->Cases (compare k_find k) of
         LESS => (coqfind left k_find)
       | EQUAL => (Some a data)
       | GREATER => (coqfind right k_find)
       end=(None a).
  Intro Z.
  Apply Z with cmp2:=(compare k_find k); Trivial.

  Intros cmp2 HO2.
  NewDestruct cmp2.
  Rewrite -> HO2.
  Apply no_find_over.
  Apply wf_order_overbound_equal with k:=x.
  Assumption.

  Inversion t1_bounded_up; Assumption.

  Rewrite -> HO2.
  Cut (compare k x)=LESS.
  Cut (compare k x)=EQUAL.
  Intro K.
  Replace (compare k x) with (compare k_find k).
  Rewrite -> HO2.
  Intro Z; Discriminate Z.

  Rewrite -> HO2.
  Rewrite -> K.
  Trivial.

  Apply compare_equal_sym.
  Replace (compare x k) with (compare k_find k).
  Rewrite -> HO2.
  Trivial.

  Apply compare_equiv.
  Assumption.

  Inversion t1_bounded_up.
  Assumption.

  Rewrite -> HO2.
  Apply no_find_over.
  Apply wf_order_overbound_equal with k:=x.
  Assumption.

  Inversion t1_bounded_up; Assumption.

  Rewrite -> HO1.
  Simpl.
  Cut
   (cmp2:compare_result)
     (compare k_find k)=cmp2
     ->Cases (compare k_find k) of
         LESS => (coqfind left k_find)
       | EQUAL => (Some a data)
       | GREATER => (coqfind right k_find)
       end=(None a).
  Intro Z.
  Apply Z with cmp2:=(compare k_find k); Trivial.

  Intros cmp2 HO2.
  NewDestruct cmp2.
  Rewrite -> HO2.
  Apply no_find_over.
  Apply wf_order_overbound with k:=x.
  Assumption.

  Inversion t1_bounded_up; Assumption.

  Rewrite -> HO2.
  Inversion t1_bounded_up.
  Cut (compare k_find x)=(compare k x).
  Intro rew.
  Cut (compare k x)=GREATER.
  Rewrite -> key_bounded.
  Intro Z; Discriminate Z.

  Replace (compare k x) with (compare k_find x).
  Assumption.

  Apply compare_equiv.
  Assumption.

  Rewrite -> HO2.
  Apply no_find_over.
  Apply wf_order_overbound with k:=x.
  Assumption.

  Inversion t1_bounded_up; Assumption.

  Intros.
  Simpl.
  Cut
   (x0:key; d:a)
     (x0,d)=(min_binding left0 k0 data0)
     ->(coqfind
         let
          (x0,d) =
           (min_binding left0 k0 data0)
         in (bal (Node left k data right height0) x0 d
              (remove_min_binding left0 k0 data0 right0)) k_find)
        =Cases (compare k_find x) of
           LESS =>
            Cases (compare k_find k) of
              LESS => (coqfind left k_find)
            | EQUAL => (Some a data)
            | GREATER => (coqfind right k_find)
            end
         | EQUAL => (None a)
         | GREATER =>
            Cases (compare k_find k0) of
              LESS => (coqfind left0 k_find)
            | EQUAL => (Some a data0)
            | GREATER => (coqfind right0 k_find)
            end
         end.
  Intro W.
  Apply
   W
     with
     x0:=(fst key a (min_binding left0 k0 data0))
    d:=(snd key a (min_binding left0 k0 data0)).
  NewDestruct (min_binding left0 k0 data0); Trivial.

  Intros x0 d Hx0d.
  Rewrite <- Hx0d.
  Cut
   (cmp1:compare_result)
     (compare k_find x0)=cmp1
     ->(coqfind
         (bal (Node left k data right height0) x0 d
           (remove_min_binding left0 k0 data0 right0)) k_find)
        =Cases (compare k_find x) of
           LESS =>
            Cases (compare k_find k) of
              LESS => (coqfind left k_find)
            | EQUAL => (Some a data)
            | GREATER => (coqfind right k_find)
            end
         | EQUAL => (None a)
         | GREATER =>
            Cases (compare k_find k0) of
              LESS => (coqfind left0 k_find)
            | EQUAL => (Some a data0)
            | GREATER => (coqfind right0 k_find)
            end
         end.
  Intro W.
  Apply W with cmp1:=(compare k_find x0); Trivial.

  Intros cmp1 HO1.
  NewDestruct cmp1.
  Rewrite -> find_bal_less.
  Cut
   (cmp2:compare_result)
     (compare k_find x)=cmp2
     ->(coqfind (Node left k data right height0) k_find)
        =Cases (compare k_find x) of
           LESS =>
            Cases (compare k_find k) of
              LESS => (coqfind left k_find)
            | EQUAL => (Some a data)
            | GREATER => (coqfind right k_find)
            end
         | EQUAL => (None a)
         | GREATER =>
            Cases (compare k_find k0) of
              LESS => (coqfind left0 k_find)
            | EQUAL => (Some a data0)
            | GREATER => (coqfind right0 k_find)
            end
         end.
  Intro Z.
  Apply Z with cmp2:=(compare k_find x); Trivial.

  Intros cmp2 HO2.
  NewDestruct cmp2.
  Rewrite -> HO2.
  Simpl.
  Trivial.

  Simpl.
  Rewrite -> HO2.
  Cut
   (cmp3:compare_result)
     (compare k_find k)=cmp3
     ->Cases (compare k_find k) of
         LESS => (coqfind left k_find)
       | EQUAL => (Some a data)
       | GREATER => (coqfind right k_find)
       end=(None a).
  Intro Z.
  Apply Z with cmp3:=(compare k_find k); Trivial.

  Intros cmp3 HO3.
  NewDestruct cmp3.
  Rewrite -> HO3.
  Apply no_find_over.
  Apply wf_order_overbound_equal with k:=x.
  Assumption.

  Inversion t1_bounded_up; Assumption.

  Rewrite -> HO3.
  Inversion t1_bounded_up.
  Cut (compare k_find x)=LESS.
  Rewrite -> HO2.
  Intro Z; Discriminate Z.

  Replace (compare k_find x) with (compare k x).
  Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Rewrite -> HO3.
  Inversion t1_bounded_up.
  Apply no_find_over.
  Apply wf_order_overbound_equal with k:=x.
  Assumption.

  Assumption.

  Rewrite -> HO2.
  Cut
   (cmp3:compare_result)
     (compare k_find k0)=cmp3
     ->(coqfind (Node left k data right height0) k_find)
        =Cases (compare k_find k0) of
           LESS => (coqfind left0 k_find)
         | EQUAL => (Some a data0)
         | GREATER => (coqfind right0 k_find)
         end.
  Intro Z.
  Apply Z with cmp3:=(compare k_find k0); Trivial.

  Intros cmp3 HO3.
  NewDestruct cmp3.
  Rewrite -> HO3.
  Simpl.
  Replace (compare k_find k) with GREATER.
  Replace (coqfind left0 k_find) with (None a).
  Replace (coqfind right k_find) with (None a).
  Trivial.

  Symmetry.
  Apply no_find_over.
  Apply wf_order_overbound with k:=x.
  Assumption.

  Inversion t1_bounded_up; Assumption.

  Symmetry.
  Apply no_find_under.
  Apply bounded_down_from_min_binding with x:=k0 d:=data0.
  Inversion Hwf_order2; Assumption.

  NewDestruct (min_binding left0 k0 data0).
  Rewrite <- Hx0d.
  Simpl.
  Assumption.

  Symmetry.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=x.
  Inversion t1_bounded_up; Assumption.

  Apply compare_reverse1; Assumption.

  Rewrite -> HO3.
  Simpl.
  Cut (compare x0 k0)=LESS\/(compare x0 k0)=EQUAL.
  Intro choice.
  Case choice.
  Clear choice.
  Intro absurd.
  Cut (compare k_find k0)=LESS.
  Intro absurd2.
  Cut (compare k0 k0)=LESS.
  Rewrite -> compare_refl.
  Intro Z; Discriminate Z.

  Apply compare_trans1 with y:=x0.
  Replace (compare k0 x0) with (compare k_find x0).
  Assumption.

  Apply compare_equiv; Assumption.

  Assumption.

  Apply compare_trans1 with y:=x0.
  Assumption.

  Assumption.

  Intro equal.
  Cut (compare k_find k0)=LESS.
  Rewrite -> HO3.
  Intro Z; Discriminate Z.

  Replace (compare k_find k0) with (compare k_find x0).
  Assumption.

  Apply compare_equiv2.
  Assumption.

  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_less_or_equal_than_node.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d.
  Simpl; Trivial.

  Simpl.
  Rewrite -> HO3.
  Cut (compare x0 k0)=LESS\/(compare x0 k0)=EQUAL.
  Intro choice.
  Case choice.
  Clear choice.
  Intro less.
  Cut (compare k_find k0)=LESS.
  Rewrite -> HO3.
  Intro absurd; Discriminate absurd.

  Apply compare_trans1 with y:=x0; Assumption.

  Intro equal.
  Cut (compare k_find k0)=LESS.
  Rewrite -> HO3.
  Intro absurd; Discriminate absurd.

  Replace (compare k_find k0) with (compare k_find x0).
  Assumption.

  Apply compare_equiv2.
  Assumption.

  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_less_or_equal_than_node.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d.
  Simpl; Trivial.

  Inversion Hwf_order2; Assumption.

  Apply remove_min_binding_wf_order.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Apply wf_order_overbound with k:=x.
  Apply compare_reverse2.
  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounded_down.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion t2_bounded_down; Assumption.

  Inversion t2_bounded_down.
  Apply compare_reverse1; Assumption.

  Rewrite <- Hx0d.
  Simpl; Trivial.

  Assumption.

  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounds_down_removed.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Assumption.

  Simpl.
  Replace
  (coqfind
    (bal (Node left k data right height0) x0 d
      (remove_min_binding left0 k0 data0 right0)) k_find) with (Some a d).
  Replace (compare k_find x) with GREATER.
  Cut
   (cmp1:compare_result)
     (compare k_find k0)=cmp1
     ->(Some a d)
        =Cases (compare k_find k0) of
           LESS => (coqfind left0 k_find)
         | EQUAL => (Some a data0)
         | GREATER => (coqfind right0 k_find)
         end.
  Intro W; Apply W with cmp1:=(compare k_find k0); Trivial.

  Intros cmp2 HO2.
  NewDestruct cmp2.
  Rewrite -> HO2.
  Cut (compare (fst key a (min_binding left0 k0 data0)) k0)=LESS.
  Replace (coqfind left0 k_find) with
  (coqfind left0 (fst key a (min_binding left0 k0 data0))).
  Replace d with (snd key a (min_binding left0 k0 data0)).
  Intro HZ.
  Symmetry.
  Apply min_binding_finds_binding1.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Assumption.

  NewDestruct (min_binding left0 k0 data0).
  Rewrite <- Hx0d; Simpl; Trivial.

  Replace (fst key a (min_binding left0 k0 data0)) with x0.
  Apply coqfind_equiv.
  Apply compare_equal_sym; Assumption.

  NewDestruct (min_binding left0 k0 data0).
  Rewrite <- Hx0d; Simpl; Trivial.

  Rewrite <- Hx0d.
  Simpl.
  Replace (compare x0 k0) with (compare k_find k0).
  Assumption.

  Apply compare_equiv.
  Assumption.

  Rewrite -> HO2.
  Replace d with data0; Trivial.
  Replace d with (snd key a (min_binding left0 k0 data0)).
  Symmetry.
  Apply min_binding_finds_binding2.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d; Simpl.
  Replace (compare x0 k0) with (compare k_find k0).
  Assumption.

  Apply compare_equiv.
  Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Rewrite -> HO2.
  Cut
   (compare (fst key a (min_binding left0 k0 data0)) k0)=LESS\/(compare
                                                                (
                                                                fst key a
                                                                (
                                                                min_binding
                                                                left0 k0
                                                                data0))
                                                                k0)=EQUAL.
  Rewrite <- Hx0d; Simpl.
  Replace (compare x0 k0) with (compare k_find k0).
  Rewrite -> HO2.
  Intro split; Case split; Intro absurd; Discriminate absurd.

  Apply compare_equiv.
  Assumption.

  Apply min_binding_less_or_equal_than_node.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Symmetry.
  Apply compare_reverse2.
  Replace (compare x k_find) with
  (compare x (fst key a (min_binding left0 k0 data0))).
  Apply min_binding_bounded_down.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion t2_bounded_down; Assumption.

  Inversion t2_bounded_down.
  Apply compare_reverse1; Assumption.

  Rewrite <- Hx0d; Simpl.
  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  Rewrite -> find_bal_imm.
  Trivial.

  Assumption.

  Apply remove_min_binding_wf_order.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Apply wf_order_overbound with k:=x.
  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply compare_reverse2.
  Apply min_binding_bounded_down.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion t2_bounded_down; Assumption.

  Inversion t2_bounded_down.
  Apply compare_reverse1.
  Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Assumption.

  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounds_down_removed.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Assumption.

  Rewrite -> find_bal_greater.
  Rewrite -> coqfind_remove_min_binding.
  Simpl.
  Replace (compare k_find x) with GREATER.
  Trivial.

  Symmetry.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=x0.
  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounded_down.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion t2_bounded_down; Assumption.

  Inversion t2_bounded_down.
  Apply compare_reverse1; Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Apply compare_reverse1.
  Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d; Simpl; Assumption.

  Assumption.

  Apply remove_min_binding_wf_order.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Apply wf_order_overbound with k:=x.
  Apply compare_reverse2.
  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounded_down.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion t2_bounded_down; Assumption.

  Inversion t2_bounded_down.
  Apply compare_reverse1.
  Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Assumption.

  Replace x0 with (fst key a (min_binding left0 k0 data0)).
  Apply min_binding_bounds_down_removed.
  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Inversion Hwf_order2; Assumption.

  Rewrite <- Hx0d; Simpl; Trivial.

  Assumption.
Qed.

