(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Fixpoint coqremove [tree : t] : key -> t := [x : key]
  Cases tree of
  | Empty => Empty
  | (Node l v d r h) =>
     Cases (compare x v) of
     | EQUAL => (merge l r)
     | LESS => (bal (coqremove l x) v d r)
     | GREATER => (bal l v d (coqremove r x))
     end
  end.

Lemma coqremove_bounded_up : (tree : t) (k_rem : key)
   (tree_wf_order : (wf_order tree))
   (y : key) (y_bounded_up : (bounded_up y tree))
   (bounded_up y (coqremove tree k_rem)).
Proof.
  Intro.
  Elim tree.
  Simpl.
  Auto.

  Intros l Hl k d r Hr h k_rem tree_wf_order y y_bounded_up.
  Inversion y_bounded_up.
  Simpl.
  Inversion tree_wf_order.
  NewDestruct (compare k_rem k).
  Apply bal_bounded_up.
  Assumption.

  Apply Hl; Assumption.

  Assumption.

  Apply merge_bounded_up with x:=k; Assumption.

  Apply bal_bounded_up.
  Assumption.

  Assumption.

  Apply Hr; Assumption.
Qed.

Lemma coqremove_bounded_down : (tree : t) (k_rem : key)
   (tree_wf_order : (wf_order tree))
   (y : key) (y_bounded_down : (bounded_down y tree))
   (bounded_down y (coqremove tree k_rem)).
Proof.
  Intro.
  Elim tree.
  Simpl.
  Auto.

  Intros l Hl k d r Hr h k_rem tree_wf_order y y_bounded_down.
  Inversion y_bounded_down.
  Simpl.
  Inversion tree_wf_order.
  NewDestruct (compare k_rem k).
  Apply bal_bounded_down.
  Assumption.

  Apply Hl; Assumption.

  Assumption.

  Apply merge_bounded_down with x:=k; Assumption.

  Apply bal_bounded_down.
  Assumption.

  Assumption.

  Apply Hr; Assumption.
Qed.

Lemma coqremove_wf_order : (tree : t) (k_rem : key)
  (tree_wf_order : (wf_order tree)) (wf_order (coqremove tree k_rem)).
Proof.
  Intro.
  Elim tree.
  Simpl.
  Auto.

  Clear tree.
  Intros l Hl k d r Hr h k_rem tree_wf_order.
  Simpl.
  Inversion tree_wf_order.
  NewDestruct (compare k_rem k).
  Apply bal_wf_order.
  Apply Hl.
  Assumption.

  Assumption.

  Apply coqremove_bounded_up; Assumption.

  Assumption.

  Apply merge_wf_order with x:=k; Assumption.

  Apply bal_wf_order.
  Assumption.

  Apply Hr; Assumption.

  Assumption.

  Apply coqremove_bounded_down; Assumption.
Qed.

Definition func_remove : (key -> (option a)) -> key -> (key -> (option a)) :=
  [f : key->(option a)] [k_rem : key] [k: key]
  Cases (compare k_rem k) of
  | EQUAL => (None a)
  | _ => (f k)
  end.

Theorem remove_correct : (ft : t) (ft_wf_order : (wf_order ft))
  (k_rem : key) (k_find : key)
  ((func_remove (coqfind ft) k_rem k_find) =
   (coqfind (coqremove ft k_rem) k_find)).
Proof.
  Intro ft.
  Elim ft.
  Unfold func_remove.
  Simpl.
  Intros.
  NewDestruct (compare k_rem k_find); Trivial.

  Intros l Hl k d r Hr h ft_wf_order k_rem k_find.
  Inversion ft_wf_order.
  Unfold func_remove.
  Simpl.
  Cut
   (cmp1:compare_result)
     (compare k_rem k_find)=cmp1
     ->Cases (compare k_rem k_find) of
         LESS =>
          Cases (compare k_find k) of
            LESS => (coqfind l k_find)
          | EQUAL => (Some a d)
          | GREATER => (coqfind r k_find)
          end
       | EQUAL => (None a)
       | GREATER =>
          Cases (compare k_find k) of
            LESS => (coqfind l k_find)
          | EQUAL => (Some a d)
          | GREATER => (coqfind r k_find)
          end
       end
        =(coqfind
           Cases (compare k_rem k) of
             LESS => (bal (coqremove l k_rem) k d r)
           | EQUAL => (merge l r)
           | GREATER => (bal l k d (coqremove r k_rem))
           end k_find).
  Intro W; Apply W with cmp1:=(compare k_rem k_find); Trivial.

  Intros cmp1 HO1.
  NewDestruct cmp1.
  Rewrite -> HO1.
  Cut
   (cmp2:compare_result)
     (compare k_find k)=cmp2
     ->Cases (compare k_find k) of
         LESS => (coqfind l k_find)
       | EQUAL => (Some a d)
       | GREATER => (coqfind r k_find)
       end
        =(coqfind
           Cases (compare k_rem k) of
             LESS => (bal (coqremove l k_rem) k d r)
           | EQUAL => (merge l r)
           | GREATER => (bal l k d (coqremove r k_rem))
           end k_find).
  Intro W; Apply W with cmp2:=(compare k_find k); Trivial.

  Intros cmp2 HO2.
  Rewrite -> HO2.
  NewDestruct cmp2.
  Replace (compare k_rem k) with LESS.
  Rewrite -> find_bal_less.
  Rewrite <- Hl.
  Unfold func_remove.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_up; Assumption.

  Assumption.

  Assumption.

  Symmetry.
  Apply compare_trans1 with y:=k_find; Assumption.

  Replace (compare k_rem k) with LESS.
  Symmetry.
  Apply find_bal_imm.
  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_up; Assumption.

  Assumption.

  Assumption.

  Replace LESS with (compare k_rem k_find).
  Apply compare_equiv2; Assumption.

  Cut
   (cmp3:compare_result)
     (compare k_rem k)=cmp3
     ->(coqfind r k_find)
        =(coqfind
           Cases (compare k_rem k) of
             LESS => (bal (coqremove l k_rem) k d r)
           | EQUAL => (merge l r)
           | GREATER => (bal l k d (coqremove r k_rem))
           end k_find).
  Intro W; Apply W with cmp3:=(compare k_rem k); Trivial.

  Intros cmp3 HO3.
  Rewrite -> HO3.
  NewDestruct cmp3.
  Rewrite -> find_bal_greater.
  Trivial.

  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_up; Assumption.

  Assumption.

  Assumption.

  Rewrite -> find_merge with x:=k.
  Rewrite -> HO2.
  Trivial.

  Assumption.

  Assumption.

  Assumption.

  Assumption.

  Rewrite -> find_bal_greater.
  Rewrite <- Hr.
  Unfold func_remove.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Assumption.

  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_down; Assumption.

  Assumption.

  Rewrite -> HO1.
  Cut
   (cmp2:compare_result)
     (compare k_rem k)=cmp2
     ->(None a)
        =(coqfind
           Cases (compare k_rem k) of
             LESS => (bal (coqremove l k_rem) k d r)
           | EQUAL => (merge l r)
           | GREATER => (bal l k d (coqremove r k_rem))
           end k_find).
  Intro W; Apply W with cmp2:=(compare k_rem k); Trivial.

  Intros cmp2 HO2.
  Rewrite -> HO2.
  NewDestruct cmp2.
  Rewrite -> find_bal_less.
  Rewrite <- Hl.
  Unfold func_remove.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_up; Assumption.

  Assumption.

  Replace (compare k_find k) with (compare k_rem k).
  Assumption.

  Apply compare_equiv; Assumption.

  Rewrite -> find_merge with x:=k.
  Replace (compare k_find k) with EQUAL.
  Trivial.

  Replace EQUAL with (compare k_rem k).
  Apply compare_equiv.
  Assumption.

  Assumption.

  Assumption.

  Assumption.

  Assumption.

  Rewrite -> find_bal_greater.
  Rewrite <- Hr.
  Unfold func_remove.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Assumption.

  Apply coqremove_wf_order.
  Assumption.

  Assumption.

  Apply coqremove_bounded_down; Assumption.

  Replace GREATER with (compare k_rem k).
  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Rewrite -> HO1.
  Cut
   (cmp2:compare_result)
     (compare k_find k)=cmp2
     ->Cases (compare k_find k) of
         LESS => (coqfind l k_find)
       | EQUAL => (Some a d)
       | GREATER => (coqfind r k_find)
       end
        =(coqfind
           Cases (compare k_rem k) of
             LESS => (bal (coqremove l k_rem) k d r)
           | EQUAL => (merge l r)
           | GREATER => (bal l k d (coqremove r k_rem))
           end k_find).
  Intro W; Apply W with cmp2:=(compare k_find k); Trivial.

  Intros cmp2 HO2; Rewrite -> HO2; NewDestruct cmp2.
  Cut
   (cmp3:compare_result)
     (compare k_rem k)=cmp3
     ->(coqfind l k_find)
        =(coqfind
           Cases (compare k_rem k) of
             LESS => (bal (coqremove l k_rem) k d r)
           | EQUAL => (merge l r)
           | GREATER => (bal l k d (coqremove r k_rem))
           end k_find).
  Intro W; Apply W with cmp3:=(compare k_rem k); Trivial.

  Intros cmp3 HO3; Rewrite -> HO3; NewDestruct cmp3.
  Rewrite -> find_bal_less.
  Rewrite <- Hl.
  Unfold func_remove.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_up; Assumption.

  Assumption.

  Assumption.

  Rewrite -> find_merge with x:=k.
  Rewrite -> HO2.
  Trivial.

  Assumption.

  Assumption.

  Assumption.

  Assumption.

  Rewrite -> find_bal_less.
  Trivial.

  Assumption.

  Apply coqremove_wf_order.
  Assumption.

  Assumption.

  Apply coqremove_bounded_down.
  Assumption.

  Assumption.

  Assumption.

  Replace (compare k_rem k) with GREATER.
  Rewrite -> find_bal_imm.
  Trivial.

  Assumption.

  Apply coqremove_wf_order.
  Assumption.

  Assumption.

  Apply coqremove_bounded_down.
  Assumption.

  Assumption.

  Assumption.

  Replace GREATER with (compare k_rem k_find).
  Apply compare_equiv2.
  Assumption.

  Replace (compare k_rem k) with GREATER.
  Rewrite -> find_bal_greater.
  Rewrite <- Hr.
  Unfold func_remove.
  Rewrite -> HO1.
  Trivial.

  Assumption.

  Assumption.

  Apply coqremove_wf_order; Assumption.

  Assumption.

  Apply coqremove_bounded_down; Assumption.

  Assumption.

  Symmetry.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=k_find; Apply compare_reverse1; Assumption.
Qed.
