(*     let rec cut v = function
        Empty -> raise Not_found
      | Node (l1,v1,d1,r1,h1) ->
          let c = Ord.compare v v1 in
          if c <0 then
            let (l2,d2,r2) = cut v l1 in (l2,d2,Node (r2,v1,d1,r1,h1))
          else if c > 0 then
            let (l2,d2,r2) = cut v r1 in (Node (l1,v1,d1,l2,h1),d2,r2)
          else (l1,d1,r1) *)


Fixpoint rotate [tree : t] : key -> t := [v : key]
  Cases tree of
  | Empty => Empty
  | (Node l1 v1 d1 r1 h1) =>
     Cases (compare v v1) of
     | LESS =>
       Cases (rotate l1 v) of
       | Empty => Empty
       | (Node l2 v2 d2 r2 h2) => (Node l2 v2 d2 (Node r2 v1 d1 r1 h2) h1)
       end
     | EQUAL => tree
     | RIGHT =>
       Cases (rotate r1 v) of
       | Empty => Empty
       | (Node l2 v2 d2 r2 h2) => (Node (Node l1 v1 d1 l2 h2) v2 d2 r2 h1)
       end
     end
  end.

Lemma rotate_bounded_up : (tree : t) (v : key) (y : key)
  (bounded_up y tree) ->
  (bounded_up y (rotate tree v)).
Proof.
  Intro tree.
  Elim tree.
  Simpl.
  Auto.

  Intros l Hl k d r Hr ht v y bounded.
  Simpl.
  NewDestruct (compare v k).
  Cut (bounded_up y (rotate l v)).
  NewDestruct (rotate l v).
  Auto.

  Intro bounded2.
  Inversion bounded.
  Inversion bounded2.
  Constructor.
  Assumption.

  Assumption.

  Constructor; Assumption.

  Apply Hl.
  Inversion bounded; Assumption.

  Assumption.

  Cut (bounded_up y (rotate r v)).
  NewDestruct (rotate r v).
  Auto.

  Intro bounded2.
  Inversion bounded.
  Inversion bounded2.
  Constructor.
  Constructor; Assumption.

  Assumption.

  Assumption.

  Apply Hr.
  Inversion bounded; Assumption.
Qed.

Lemma rotate_bounded_down : (tree : t) (v : key) (y : key)
  (bounded_down y tree) ->
  (bounded_down y (rotate tree v)).
Proof.
  Intro tree.
  Elim tree.
  Simpl.
  Auto.

  Intros l Hl k d r Hr ht v y bounded.
  Simpl.
  NewDestruct (compare v k).
  Cut (bounded_down y (rotate l v)).
  NewDestruct (rotate l v).
  Auto.

  Intro bounded2.
  Inversion bounded.
  Inversion bounded2.
  Constructor.
  Assumption.

  Assumption.

  Constructor; Assumption.

  Apply Hl.
  Inversion bounded; Assumption.

  Assumption.

  Cut (bounded_down y (rotate r v)).
  NewDestruct (rotate r v).
  Auto.

  Intro bounded2.
  Inversion bounded.
  Inversion bounded2.
  Constructor.
  Constructor; Assumption.

  Assumption.

  Assumption.

  Apply Hr.
  Inversion bounded; Assumption.
Qed.

Lemma rotate_returns_center : (tree : t) (v : key)
  (Cases (rotate tree v) of
   | Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end).
Proof.
  Intro tree.
  Elim tree.
  Simpl.
  Auto.

  Intros l Hl k d r Hr h.
  Simpl.
  Intro v.
  Cut
   (cmp:compare_result)
     (compare v k)=cmp
     ->Cases
         Cases (compare v k) of
           LESS =>
            Cases (rotate l v) of
              Empty => Empty
            | (Node l2 v2 d2 r2 h2) =>
               (Node l2 v2 d2 (Node r2 k d r h2) h)
            end
         | EQUAL => (Node l k d r h)
         | GREATER =>
            Cases (rotate r v) of
              Empty => Empty
            | (Node l2 v2 d2 r2 h2) =>
               (Node (Node l k d l2 h2) v2 d2 r2 h)
            end
         end
       of
         Empty => True
       | (Node _ k0 _ _ _) => (compare v k0)=EQUAL
       end.
  Intro W; Apply W with cmp:=(compare v k); Trivial.

  Intros cmp Hcompare.
  Rewrite -> Hcompare.
  NewDestruct cmp.
  Cut
   Cases (rotate l v) of
     Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end.
  NewDestruct (rotate l v).
  Auto.

  Auto.

  Apply Hl.

  Assumption.

  Cut
   Cases (rotate r v) of
     Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end.
  NewDestruct (rotate r v).
  Auto.

  Auto.

  Apply Hr.
Qed.

Lemma  rotate_wf_order : (tree : t) (v : key)
  (wf_order tree) ->
  (wf_order (rotate tree v)).
Proof.
  Intro.
  Elim tree.
  Simpl; Auto.

  Intros l Hl k d r Hr height0 v tree_wf_order.
  Inversion tree_wf_order.
  Simpl.
  TCompare v k Horder.
  Cut (wf_order (rotate l v)).
  Cut (bounded_up k (rotate l v)).
  Intros Hbounded1 Hwf_order1.
  NewDestruct (rotate l v).
  Trivial.

  Inversion Hbounded1.
  Inversion Hwf_order1.
  Constructor; Try Assumption.
  Constructor; Try Assumption.

  Constructor; Try Assumption.
  Apply compare_reverse2; Assumption.

  Apply wf_order_underbound with k:=k; Try Assumption.

  Apply rotate_bounded_up; Assumption.

  Apply Hl; Assumption.

  Constructor; Try Assumption.

  Cut (bounded_down k (rotate r v)).
  Cut (wf_order (rotate r v)).
  Intros Hwf_order1 Hbounded1.
  NewDestruct (rotate r v).
  Trivial.

  Inversion Hwf_order1.
  Inversion Hbounded1.
  Constructor; Try Assumption.
  Constructor; Try Assumption.

  Constructor; Try Assumption.
  Apply wf_order_overbound with k:=k; Assumption.

  Apply compare_reverse1; Assumption.

  Apply Hr; Assumption.

  Apply rotate_bounded_down; Assumption.
Qed.

Lemma rotate_v_find : (tree : t) (v : key)
  (Cases (coqfind tree v) of
   | None => True
   | (Some d) => (coqfind (rotate tree v) v) = (Some a d)
   end).
Proof.
  Intros tree v.
  Elim tree.
  Simpl.
  Trivial.

  Intros l Hl k d r Hr height0.
  Simpl.
  TCompare v k Hvk.
  NewDestruct (coqfind l v).
  Cut (coqfind (rotate l v) v)=(Some a a0).
  NewDestruct (rotate l v).
  Simpl.
  Auto.

  Simpl.
  TCompare v k0 Hvk0.
  Auto.

  Auto.

  Rewrite -> Hvk.
  Auto.

  Apply Hl.

  Trivial.

  Simpl.
  Rewrite -> Hvk.
  Trivial.

  NewDestruct (coqfind r v).
  Cut (coqfind (rotate r v) v)=(Some a a0).
  NewDestruct (rotate r v).
  Auto.

  Simpl.
  Rewrite -> Hvk.
  Trivial.

  Assumption.

  Trivial.
Qed.

Lemma rotate_not_empty : (tree : t) (v : key)
  (Cases (coqfind tree v) of
   | None => True
   | (Some d) => (rotate tree v)=Empty -> False
   end).
Proof.
  Intros.
  Cut
   Cases (coqfind tree v) of
     None => True
   | (Some d) => (coqfind (rotate tree v) v)=(Some a d)
   end.
  NewDestruct (coqfind tree v).
  Intros H1 H2.
  Cut (coqfind (rotate tree v) v)=(Some a a0).
  Rewrite -> H2.
  Simpl.
  Intro Z.
  Discriminate Z.

  Assumption.

  Auto.

  Apply rotate_v_find.
Qed.

Lemma  rotate_correct : (tree : t) (v : key) (k_find : key)
  (tree_wf_order : (wf_order tree))
  (Cases (coqfind tree v) of
   | None => True
   | (Some _) => (coqfind (rotate tree v) k_find) = (coqfind tree k_find)
   end).

Proof.
  Intros.
  Elim tree.
  Simpl; Trivial.

  Intros l Hl k d r Hr height0.
  Simpl.
  TCompare v k Hvk.
  Cut
   (vimage:(option a))
     (coqfind l v)=vimage
     ->Cases (coqfind l v) of
         (Some _) =>
          (coqfind
            Cases (rotate l v) of
              Empty => Empty
            | (Node l2 v2 d2 r2 h2) =>
               (Node l2 v2 d2 (Node r2 k d r h2) height0)
            end k_find)
           =Cases (compare k_find k) of
              LESS => (coqfind l k_find)
            | EQUAL => (Some a d)
            | GREATER => (coqfind r k_find)
            end
       | None => True
       end.
  Intro W; Apply W with vimage:=(coqfind l v); Trivial.

  Intros vimage Rlvimage.
  Rewrite -> Rlvimage.
  NewDestruct vimage.
  TCompare k_find k Hfind_k.
  Cut (coqfind (rotate l v) k_find)=(coqfind l k_find).
  Intro brew.
  Rewrite <- brew.
  NewDestruct (rotate l v).
  Trivial.

  Simpl.
  Rewrite -> Hfind_k.
  Trivial.

  Cut
   Cases (coqfind l v) of
     (Some _) => (coqfind (rotate l v) k_find)=(coqfind l k_find)
   | None => True
   end.
  Rewrite -> Rlvimage.
  Auto.

  Apply Hl.

  Cut
   Cases (coqfind l v) of
     (Some _) => (coqfind (rotate l v) k_find)=(coqfind l k_find)
   | None => True
   end.
  Rewrite -> Rlvimage.
  Cut
   (rlv:t)
     (rotate l v)=rlv
     ->(coqfind (rotate l v) k_find)=(coqfind l k_find)
     ->(coqfind
         Cases (rotate l v) of
           Empty => Empty
         | (Node l2 v2 d2 r2 h2) =>
            (Node l2 v2 d2 (Node r2 k d r h2) height0)
         end k_find)=(Some a d).
  Intro W; Apply W with rlv:=(rotate l v); Trivial.

  Intros rlv Hrlv.
  Rewrite -> Hrlv.
  NewDestruct rlv.
  Apply False_ind.
  Cut
   Cases (coqfind l v) of
     None => True
   | (Some d) => (rotate l v)=Empty->False
   end.
  Rewrite -> Rlvimage.
  Rewrite -> Hrlv.
  Auto.

  Apply rotate_not_empty.

  Intro foo.
  Simpl.
  Rewrite -> Hfind_k.
  Cut (compare v k0)=EQUAL.
  Intro Hvk0.
  Replace (compare k_find k0) with (compare k_find v).
  Replace (compare k_find v) with (compare k v).
  Replace (compare k v) with GREATER.
  Trivial.

  Symmetry.
  Apply compare_reverse2; Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym; Assumption.

  Apply compare_equiv2; Assumption.

  Cut
   Cases (rotate l v) of
     Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end.
  Rewrite -> Hrlv.
  Auto.

  Apply rotate_returns_center.

  Apply Hl.

  Cut
   Cases (coqfind l v) of
     None => True
   | (Some d) => (rotate l v)=Empty->False
   end.
  Cut
   (tree:t)
     (rotate l v)=tree
     ->Cases (coqfind l v) of
         (Some _) => (rotate l v)=Empty->False
       | None => True
       end
     ->(coqfind
         Cases (rotate l v) of
           Empty => Empty
         | (Node l2 v2 d2 r2 h2) =>
            (Node l2 v2 d2 (Node r2 k d r h2) height0)
         end k_find)=(coqfind r k_find).
  Intro W; Apply W with tree:=(rotate l v); Trivial.

  Intros rlv Hrlv.
  Rewrite -> Hrlv.
  NewDestruct rlv.
  Rewrite -> Rlvimage.
  Tauto.

  Rewrite -> Rlvimage.
  Intro Z; Clear Z.
  Simpl.
  Rewrite -> Hfind_k.
  Replace (compare k_find k0) with GREATER.
  Trivial.

  Symmetry; Apply compare_reverse2.
  Apply compare_trans1 with y:=k.
  Replace LESS with (compare v k).
  Apply compare_equiv.
  Cut
   Cases (rotate l v) of
     Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end.
  Rewrite -> Hrlv.
  Intro eq.
  Apply compare_equal_sym.
  Assumption.

  Apply rotate_returns_center.

  Apply compare_reverse1; Assumption.

  Apply rotate_not_empty.

  Trivial.

  Simpl.
  Trivial.

  Cut
   (vimage:(option a))
     (coqfind r v)=vimage
     ->Cases (coqfind r v) of
         (Some _) =>
          (coqfind
            Cases (rotate r v) of
              Empty => Empty
            | (Node l2 v2 d2 r2 h2) =>
               (Node (Node l k d l2 h2) v2 d2 r2 height0)
            end k_find)
           =Cases (compare k_find k) of
              LESS => (coqfind l k_find)
            | EQUAL => (Some a d)
            | GREATER => (coqfind r k_find)
            end
       | None => True
       end.
  Intro W; Apply W with vimage:=(coqfind r v); Trivial.

  Intros vimage Rlvimage.
  Rewrite -> Rlvimage.
  NewDestruct vimage.
  TCompare k_find k Hfind_k.
  Cut
   (rv:t)
     (rotate r v)=rv
     ->(coqfind
         Cases (rotate r v) of
           Empty => Empty
         | (Node l2 v2 d2 r2 h2) =>
            (Node (Node l k d l2 h2) v2 d2 r2 height0)
         end k_find)=(coqfind l k_find).
  Intro W; Apply W with rv:=(rotate r v); Trivial.

  Intros rv Hrv.
  Rewrite -> Hrv.
  NewDestruct rv.
  Cut
   Cases (coqfind r v) of
     None => True
   | (Some d) => (rotate r v)=Empty->False
   end.
  Rewrite -> Rlvimage.
  Rewrite -> Hrv.
  Tauto.

  Apply rotate_not_empty.

  Simpl.
  Rewrite -> Hfind_k.
  Replace (compare k_find k0) with LESS.
  Trivial.

  Symmetry.
  Apply compare_trans1 with y:=k.
  Assumption.

  Replace LESS with (compare k v).
  Apply compare_equiv2.
  Cut
   Cases (rotate r v) of
     Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end.
  Rewrite -> Hrv.
  Intro Z.
  Apply compare_equal_sym.
  Assumption.

  Apply rotate_returns_center.

  Apply compare_reverse1; Assumption.

  Cut
   (rv:t)
     (rotate r v)=rv
     ->(coqfind
         Cases (rotate r v) of
           Empty => Empty
         | (Node l2 v2 d2 r2 h2) =>
            (Node (Node l k d l2 h2) v2 d2 r2 height0)
         end k_find)=(Some a d).
  Intro W; Apply W with rv:=(rotate r v); Trivial.

  Intros rv Hrv.
  Rewrite -> Hrv.
  NewDestruct rv.
  Cut
   Cases (coqfind r v) of
     None => True
   | (Some d) => (rotate r v)=Empty->False
   end.
  Rewrite -> Hrv.
  Rewrite -> Rlvimage.
  Tauto.

  Apply rotate_not_empty.

  Simpl.
  Replace (compare k_find k0) with (compare k_find v).
  Rewrite -> Hfind_k.
  Replace (compare k_find v) with (compare k v).
  Replace (compare k v) with LESS.
  Trivial.

  Symmetry.
  Apply compare_reverse1; Assumption.

  Apply compare_equiv.
  Apply compare_equal_sym.
  Assumption.

  Apply compare_equiv2.
  Cut
   Cases (rotate r v) of
     Empty => True
   | (Node _ k _ _ _) => (compare v k)=EQUAL
   end.
  Rewrite -> Hrv.
  Auto.

  Apply rotate_returns_center.

  Cut
   (tree:t)
     (rotate r v)=tree
     ->(coqfind
         Cases (rotate r v) of
           Empty => Empty
         | (Node l2 v2 d2 r2 h2) =>
            (Node (Node l k d l2 h2) v2 d2 r2 height0)
         end k_find)=(coqfind r k_find).
  Intro W; Apply W with tree:=(rotate r v); Trivial.

  Intros rv Hrv.
  Rewrite -> Hrv.
  NewDestruct rv.
  Cut
   Cases (coqfind r v) of
     None => True
   | (Some d) => (rotate r v)=Empty->False
   end.
  Rewrite -> Rlvimage.
  Rewrite -> Hrv.
  Tauto.

  Apply rotate_not_empty.

  Simpl.
  Rewrite -> Hfind_k.
  Cut
   Cases (coqfind r v) of
     (Some _) => (coqfind (rotate r v) k_find)=(coqfind r k_find)
   | None => True
   end.
  Rewrite -> Rlvimage.
  Rewrite -> Hrv.
  Simpl.
  Trivial.

  Apply Hr.

  Trivial.
Qed.
